# -*- coding: utf-8 -*-
"""SQLite parser plugin for iOS datausage.sqlite database files."""

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class IOSDatausageEventData(events.EventData):
  """iOS datausage event data.

  Attributes:
    bundle_identifier (str): bundle identifier.
    process_name (str): name of the process.
    start_time (dfdatetime.DateTimeValues): date and time the start of
        the network connection was established.
    wifi_in (int): number of bytes received over Wi-Fi.
    wifi_out (int): number of bytes sent over Wi-Fi.
    wireless_wan_in (int): number of bytes received over cellular.
    wireless_wan_out (int): number of bytes sent over cellular.
  """
  DATA_TYPE = 'ios:datausage:event'

  def __init__(self):
    """Initializes event data."""
    super(IOSDatausageEventData, self).__init__(data_type=self.DATA_TYPE)
    self.bundle_identifier = None
    self.process_name = None
    self.start_time = None
    self.wifi_in = None
    self.wifi_out = None
    self.wireless_wan_in = None
    self.wireless_wan_out = None


class IOSDatausagePlugin(interface.SQLitePlugin):
  """SQLite parser plugin for iOS DataUsage database."""

  NAME = "ios_datausage"
  DATA_FORMAT = 'iOS data usage SQLite databse (DataUsage.sqlite) file.'

  REQUIRED_STRUCTURE = {
      'ZLIVEUSAGE': frozenset([
          'Z_PK', 'Z_ENT', 'Z_OPT', 'ZKIND', 'ZMETADATA', 'ZTAG', 'ZHASPROCESS',
          'ZTIMESTAMP', 'ZWIFIIN', 'ZWIFIOUT', 'ZWWANIN', 'ZWWANOUT']),
      'ZPROCESS': frozenset([
          'Z_PK', 'Z_ENT', 'Z_OPT', 'ZFIRSTTIMESTAMP', 'ZTIMESTAMP',
          'ZBUNDLENAME', 'ZPROCNAME'])}

  QUERIES = [(
      ('SELECT ZLIVEUSAGE.ZTIMESTAMP, ZLIVEUSAGE.ZWIFIIN, ZLIVEUSAGE.ZWIFIOUT, '
       'ZLIVEUSAGE.ZWWANIN, ZLIVEUSAGE.ZWWANOUT, ZPROCESS.ZBUNDLENAME, '
       'ZPROCESS.ZPROCNAME FROM ZLIVEUSAGE '
       'LEFT JOIN ZPROCESS ON ZPROCESS.Z_PK = ZLIVEUSAGE.ZHASPROCESS'),
     'ParseDatausageEventRow')]

  SCHEMAS = {
    'ZLIVEUSAGE': (
        'CREATE TABLE ZLIVEUSAGE ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
        'Z_OPT INTEGER, ZKIND INTEGER, ZMETADATA INTEGER, ZTAG INTEGER, '
        'ZHASPROCESS INTEGER, ZBILLCYCLEEND TIMESTAMP, ZTIMESTAMP TIMESTAMP, '
        'ZWIFIIN FLOAT, ZWIFIOUT FLOAT, ZWWANIN FLOAT, ZWWANOUT FLOAT )'),
    'ZPROCESS': (
        'CREATE TABLE ZPROCESS ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
        'Z_OPT INTEGER, ZFIRSTTIMESTAMP TIMESTAMP, ZTIMESTAMP TIMESTAMP, '
        'ZBUNDLENAME VARCHAR, ZPROCNAME VARCHAR )')}

  REQUIRES_SCHEMA_MATCH = False

  def _GetTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.CocoaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)

  # pylint: disable=unused-argument
  def ParseDatausageEventRow(
        self, parser_mediator, query, row, **unused_kwargs):
    """Parses a row from the Datausage sqlite file.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = IOSDatausageEventData()
    event_data.bundle_identifier = self._GetRowValue(
        query_hash, row, 'ZBUNDLENAME')
    event_data.process_name = self._GetRowValue(self, row, 'ZPROCNAME')
    event_data.start_time = self._GetTimeRowValue(self, row, 'ZTIMESTAMP')
    event_data.wifi_in = int(self._GetRowValue(self, row, 'ZWIFIIN'))
    event_data.wifi_out = int(self._GetRowValue(self, row, 'ZWIFIOUT'))
    event_data.wireless_wan_in = int(self._GetRowValue(self, row, 'ZWWANIN'))
    event_data.wireless_wan_out = int(self._GetRowValue(self, row, 'ZWWANOUT'))

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(IOSDatausagePlugin)



# -*- coding: utf-8 -*-
"""Text parser plugin for MacOS security daemon (securityd) log files.

Also see:
  https://opensource.apple.com/source/Security/Security-55471/sec/securityd
"""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import dateless_helper
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class MacOSSecuritydLogEventData(events.EventData):
  """MacOS securityd log event data.

  Attributes:
    added_time (dfdatetime.DateTimeValues): date and time the log entry
        was added.
    caller (str): caller, consists of two hex numbers.
    facility (str): facility.
    level (str): priority level.
    message (str): message.
    security_api (str): name of securityd function.
    sender (str): name of the sender.
    sender_pid (int): process identifier of the sender.
  """

  DATA_TYPE = 'macos:securityd_log:entry'

  def __init__(self):
    """Initializes event data."""
    super(MacOSSecuritydLogEventData, self).__init__(data_type=self.DATA_TYPE)
    self.added_time = None
    self.caller = None
    self.facility = None
    self.level = None
    self.message = None
    self.security_api = None
    self.sender = None
    self.sender_pid = None


class MacOSSecuritydLogTextPlugin(
    interface.TextPlugin, dateless_helper.DateLessLogFormatHelper):
  """Text parser plugin for MacOS security daemon (securityd) log files."""

  NAME = 'mac_securityd'
  DATA_FORMAT = 'MacOS security daemon (securityd) log file'

  ENCODING = 'utf-8'

  _INTEGER = pyparsing.Word(pyparsing.nums).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _ONE_OR_TWO_DIGITS = pyparsing.Word(pyparsing.nums, max=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _THREE_LETTERS = pyparsing.Word(pyparsing.alphas, exact=3)

  _DATE_TIME = pyparsing.Group(
      _THREE_LETTERS.set_results_name('month') +
      _ONE_OR_TWO_DIGITS.set_results_name('day') +
      _TWO_DIGITS.set_results_name('hours') + pyparsing.Suppress(':') +
      _TWO_DIGITS.set_results_name('minutes') + pyparsing.Suppress(':') +
      _TWO_DIGITS.set_results_name('seconds'))

  _PROCESS_IDENTIFIER = pyparsing.Word(pyparsing.nums, max=5).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _LOG_LINE = (
      _DATE_TIME.set_results_name('date_time') +
      pyparsing.CharsNotIn('[').set_results_name('sender') +
      pyparsing.Suppress('[') +
      _PROCESS_IDENTIFIER.set_results_name('sender_pid') +
      pyparsing.Suppress(']') +
      pyparsing.Suppress('<') +
      pyparsing.CharsNotIn('>').set_results_name('level') +
      pyparsing.Suppress('>') +
      pyparsing.Suppress('[') +
      pyparsing.CharsNotIn('{').set_results_name('facility') +
      pyparsing.Suppress('{') +
      pyparsing.Optional(pyparsing.CharsNotIn(
          '}').set_results_name('security_api')) +
      pyparsing.Suppress('}') +
      pyparsing.Optional(pyparsing.CharsNotIn(']:').set_results_name(
          'caller')) + pyparsing.Suppress(']:') +
      pyparsing.restOfLine().set_results_name('message') +
      _END_OF_LINE)

  _REPEATED_LOG_LINE = (
      _DATE_TIME.set_results_name('date_time') +
      pyparsing.Suppress('--- last message repeated') +
      _INTEGER.set_results_name('times') +
      pyparsing.Suppress('time ---') +
      _END_OF_LINE)

  _LINE_STRUCTURES = [
      ('log_liine', _LOG_LINE),
      ('repeated_log_line', _REPEATED_LOG_LINE)]

  VERIFICATION_GRAMMAR = _LOG_LINE

  def __init__(self):
    """Initializes a text parser plugin."""
    super(MacOSSecuritydLogTextPlugin, self).__init__()
    self._repeated_structure = None

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    # If the actual entry is a repeated entry, we take the basic information
    # from the previous entry, but use the timestamp from the actual entry.
    if key == 'log_liine':
      self._repeated_structure = structure

      message = self._GetStringValueFromStructure(structure, 'message')
    else:
      repeat_count = self._GetValueFromStructure(structure, 'times')

      structure = self._repeated_structure

      message = self._GetStringValueFromStructure(structure, 'message')
      message = 'Repeated {0:d} times: {1:s}'.format(repeat_count, message)

    event_data = MacOSSecuritydLogEventData()
    event_data.added_time = self._ParseTimeElements(time_elements_structure)
    event_data.caller = self._GetStringValueFromStructure(structure, 'caller')
    event_data.facility = self._GetValueFromStructure(structure, 'facility')
    event_data.level = self._GetValueFromStructure(structure, 'level')
    event_data.message = message or None
    event_data.security_api = self._GetValueFromStructure(
        structure, 'security_api')
    event_data.sender_pid = self._GetValueFromStructure(structure, 'sender_pid')
    event_data.sender = self._GetStringValueFromStructure(structure, 'sender')

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      month_string, day_of_month, hours, minutes, seconds = (
          time_elements_structure)

      month = self._GetMonthFromString(month_string)

      self._UpdateYear(month)

      relative_year = self._GetRelativeYear()

      time_elements_tuple = (
          relative_year, month, day_of_month, hours, minutes, seconds)

      return dfdatetime_time_elements.TimeElements(
          is_delta=True, time_elements_tuple=time_elements_tuple)

    except (TypeError, ValueError) as exception:
      raise errors.ParseError(
          'Unable to parse time elements with error: {0!s}'.format(exception))

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    self._SetEstimatedYear(parser_mediator)

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    self._repeated_structure = None

    return True


text_parser.TextLogParser.RegisterPlugin(MacOSSecuritydLogTextPlugin)



# -*- coding: utf-8 -*-
"""Plist parser plugin for Mac OS login window plist files."""

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class MacOSLoginWindowEventData(events.EventData):
  """Mac OS login window event data.

  Also see:
  * https://developer.apple.com/documentation/devicemanagement/loginwindow
  * https://developer.apple.com/documentation/devicemanagement/
    loginwindowscripts

  Attributes:
    login_hook (str): path of the script to run during login.
    logout_hook (str): path of the script to run during logout.
  """

  DATA_TYPE = 'macos:login_window:entry'

  def __init__(self):
    """Initializes event data."""
    super(MacOSLoginWindowEventData, self).__init__(data_type=self.DATA_TYPE)
    self.login_hook = None
    self.logout_hook = None


class MacOSLoginWindowManagedLoginItemEventData(events.EventData):
  """Mac OS login window managed login item event data.

  Also see:
  * https://developer.apple.com/documentation/devicemanagement/
    loginitemsmanageditems/loginitem

  Attributes:
    is_hidden (bool): True if the item should is not shown in the "Users &
        Groups" items list.
    path (str): URL or path of the location of the item.
  """

  DATA_TYPE = 'macos:login_window:managed_login_item'

  def __init__(self):
    """Initializes event data."""
    super(MacOSLoginWindowManagedLoginItemEventData, self).__init__(
        data_type=self.DATA_TYPE)
    self.is_hidden = None
    self.path = None


class MacOSLoginWindowPlugin(interface.PlistPlugin):
  """Plist parser plugin for Mac OS login window plist files."""

  NAME = 'macos_login_window_plist'
  DATA_FORMAT = 'Mac OS login window plist file'

  PLIST_PATH_FILTERS = frozenset([
    interface.PlistPathFilter('loginwindow.plist'),
    interface.PlistPathFilter('com.apple.loginwindow.plist')])

  PLIST_KEYS = frozenset([])

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, top_level=None, **unused_kwargs):
    """Extracts login window information from the plist.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      top_level (Optional[dict[str, object]]): plist top-level item.
    """
    for item in top_level.get('AutoLaunchedApplicationDictionary', []):
      event_data = MacOSLoginWindowManagedLoginItemEventData()
      event_data.is_hidden = item.get('Hide')
      event_data.path = item.get('Path')
      parser_mediator.ProduceEventData(event_data)

    event_data = MacOSLoginWindowEventData()
    event_data.login_hook = top_level.get('LoginHook')
    event_data.logout_hook = top_level.get('LogoutHook')
    parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(MacOSLoginWindowPlugin)



# -*- coding: utf-8 -*-
"""Tests for MacOS notes plugin."""

import unittest

from plaso.parsers.sqlite_plugins import macos_notes

from tests.parsers.sqlite_plugins import test_lib


class MacOSNotesTest(test_lib.SQLitePluginTestCase):
  """Tests for MacOS notes database plugin."""

  def testProcess(self):
    """Test the Process function on a Mac Notes file."""
    plugin_object = macos_notes.MacOSNotesPlugin()
    storage_writer = self._ParseDatabaseFileWithPlugin(
        ['NotesV7.storedata'], plugin_object)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 3)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'creation_time': '2014-02-11T02:38:27.097813+00:00',
        'data_type': 'macos:notes:entry',
        'modification_time': '2015-07-31T19:05:46.372972+00:00',
        'text': (
            'building 4th brandy gibs microsoft office body soul and peace '
            'example.com 3015555555: plumbing and heating claim#123456 Small '
            'business '),
        'title': 'building 4th brandy gibs'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Parser for custom destinations jump list (.customDestinations-ms) files."""

import os

from dfvfs.lib import definitions
from dfvfs.lib import errors as dfvfs_errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import resolver

from plaso.lib import dtfabric_helper
from plaso.lib import errors
from plaso.lib import specification
from plaso.parsers import interface
from plaso.parsers import manager
from plaso.parsers import winlnk


class CustomDestinationsParser(
    interface.FileObjectParser, dtfabric_helper.DtFabricHelper):
  """Parses custom destinations jump list (.customDestinations-ms) files."""

  NAME = 'custom_destinations'
  DATA_FORMAT = 'Custom destinations jump list (.customDestinations-ms) file'

  _INITIAL_FILE_OFFSET = None

  _DEFINITION_FILE = os.path.join(
      os.path.dirname(__file__), 'custom_destinations.yaml')

  # We cannot use the parser registry here since winlnk could be disabled.
  # TODO: see if there is a more elegant solution for this.
  _WINLNK_PARSER = winlnk.WinLnkParser()

  _CATEGORY_FOOTER_SIGNATURE = b'\xab\xfb\xbf\xba'

  _LNK_GUID = (
      b'\x01\x14\x02\x00\x00\x00\x00\x00\xc0\x00\x00\x00\x00\x00\x00\x46')

  def _ParseCategoryHeader(self, file_object, file_offset):
    """Parse a category header.

    Args:
      file_object (file): file-like object.
      file_offset (int): offset of the category header relative to the start of
          the file.

    Returns:
      tuple[custom_category_header, int]: category header and the number of
          bytes read.

    Raises:
      ParseError: if the category header cannot be read.
    """
    data_type_map = self._GetDataTypeMap('custom_category_header')

    category_header, bytes_read = self._ReadStructureFromFileObject(
        file_object, file_offset, data_type_map)

    if category_header.category_type > 2:
      raise errors.ParseError(
          f'Unsupported category type: {category_header.category_type:d}.')

    file_offset += bytes_read
    total_bytes_read = bytes_read

    data_type_map = self._GetDataTypeMap(
        f'custom_category_header_type_{category_header.category_type:d}')

    category_header_value, bytes_read = self._ReadStructureFromFileObject(
        file_object, file_offset, data_type_map)

    if category_header.category_type in (0, 2):
      setattr(category_header, 'number_of_entries',
              category_header_value.number_of_entries)

    total_bytes_read += bytes_read

    return category_header, total_bytes_read

  def _ParseLNKFile(
      self, parser_mediator, file_entry, file_offset, remaining_file_size):
    """Parses a LNK file stored within the .customDestinations-ms file.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      file_entry (dfvfs.FileEntry): a file entry.
      file_offset (int): offset to the LNK file, relative to the start of the
          .customDestinations-ms file.
      remaining_file_size (int): size of the data remaining in the
          .customDestinations-ms file.

    Returns:
      int: size of the LNK file data or 0 if the LNK file could not be read.
    """
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_DATA_RANGE, range_offset=file_offset,
        range_size=remaining_file_size, parent=file_entry.path_spec)

    display_name = f'{file_entry.name:s} # 0x{file_offset:08x}'

    try:
      lnk_file_object = resolver.Resolver.OpenFileObject(
          path_spec, resolver_context=parser_mediator.resolver_context)
    except (dfvfs_errors.BackEndError, RuntimeError) as exception:
      parser_mediator.ProduceExtractionWarning((
          f'unable to open LNK file: {display_name:s} with error: '
          f'{exception!s}'))
      return 0

    parser_mediator.AppendToParserChain(self._WINLNK_PARSER.NAME)
    try:
      lnk_file_object.seek(0, os.SEEK_SET)
      self._WINLNK_PARSER.ParseFileLNKFile(
          parser_mediator, lnk_file_object, display_name)
    finally:
      parser_mediator.PopFromParserChain()

    # We cannot trust the file size in the LNK data so we get the last offset
    # that was read instead.
    return lnk_file_object.get_offset()

  @classmethod
  def GetFormatSpecification(cls):
    """Retrieves the format specification.

    Returns:
      FormatSpecification: format specification.
    """
    format_specification = specification.FormatSpecification(cls.NAME)
    format_specification.AddNewSignature(b'\xab\xfb\xbf\xba', offset=-4)
    return format_specification

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a .customDestinations-ms file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      file_object (dfvfs.FileIO): a file-like object.

    Raises:
      WrongParser: when the file cannot be parsed.
    """
    file_entry = parser_mediator.GetFileEntry()
    display_name = parser_mediator.GetDisplayName()
    file_size = file_object.get_size()

    file_header_map = self._GetDataTypeMap('custom_file_header')

    try:
      file_header, file_offset = self._ReadStructureFromFileObject(
          file_object, 0, file_header_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.WrongParser((
          f'Invalid Custom Destination: {display_name:s} - unable to parse '
          f'file header with error: {exception!s}'))

    entry_header_data_type_map = self._GetDataTypeMap('custom_entry_header')
    category_footer_data_type_map = self._GetDataTypeMap(
        'custom_category_footer')

    for _ in range(file_header.number_of_categories):
      try:
        category_header, bytes_read = self._ParseCategoryHeader(
            file_object, file_offset)
      except (ValueError, errors.ParseError) as exception:
        raise errors.WrongParser((
            'Invalid Custom Destination: {display_name:s} - unable to parse '
            f'category header with error: {exception!s}'))

      file_offset += bytes_read

      number_of_entries = getattr(category_header, 'number_of_entries', 0)
      for entry_index in range(number_of_entries):
        if file_size - file_offset < 16:
          break

        try:
          entry_header, _ = self._ReadStructureFromFileObject(
              file_object, file_offset, entry_header_data_type_map)

        except errors.ParseError as exception:
          parser_mediator.ProduceExtractionWarning((
              f'unable to parse entry header at offset: 0x{file_offset:08x} '
              f'with error: {exception!s}'))
          break

        if entry_header.guid == self._LNK_GUID:
          file_offset += 16

          remaining_file_size = file_size - file_offset

          lnk_file_size = self._ParseLNKFile(
              parser_mediator, file_entry, file_offset, remaining_file_size)

          file_offset += lnk_file_size

        elif entry_header.guid[:4] != self._CATEGORY_FOOTER_SIGNATURE:
          parser_mediator.ProduceExtractionWarning((
              f'unsupported entry: {entry_index:d} at offset: '
              f'0x{file_offset:08x}'))
          break

        file_object.seek(file_offset, os.SEEK_SET)

      try:
        _, bytes_read = self._ReadStructureFromFileObject(
            file_object, file_offset, category_footer_data_type_map)
      except (ValueError, errors.ParseError) as exception:
        raise errors.WrongParser((
            f'Invalid Custom Destination: {display_name:s} - unable to parse '
            f'category footer with error: {exception!s}'))

      file_offset += 4


manager.ParsersManager.RegisterParser(CustomDestinationsParser)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Tests for Android logcat text parser plugin."""

import unittest

from dfvfs.helpers import fake_file_system_builder

from plaso.parsers import text_parser
from plaso.parsers.text_plugins import android_logcat

from tests.parsers.text_plugins import test_lib


class AndroidLogcatTextPluginTest(test_lib.TextPluginTestCase):
  """Tests for Android logcat text parser plugin."""

  def testCheckRequiredFormat(self):
    """Tests for the CheckRequiredFormat method."""
    plugin = android_logcat.AndroidLogcatTextPlugin()

    file_system_builder = fake_file_system_builder.FakeFileSystemBuilder()
    file_system_builder.AddFile('/file.txt', (
        b'--------- beginning of main\n'))

    file_entry = file_system_builder.file_system.GetFileEntryByPath('/file.txt')

    parser_mediator = self._CreateParserMediator(None, file_entry=file_entry)

    file_object = file_entry.GetFileObject()
    text_reader = text_parser.EncodedTextReader(file_object)
    text_reader.ReadLines()

    result = plugin.CheckRequiredFormat(parser_mediator, text_reader)
    self.assertTrue(result)

  def testProcess(self):
    """Tests the Process function."""
    plugin = android_logcat.AndroidLogcatTextPlugin()
    storage_writer = self._ParseTextFileWithPlugin(
        ['android_logcat.log'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 8)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    # Check entry with recorded time without a year with milliseconds precision
    # in local time.
    expected_event_values = {
        'component_tag': 'threadtime',
        'data_type': 'android:logcat',
        'message': 'test of default threadtime format',
        'pid': 1234,
        'priority': 'D',
        'recorded_time': '0000-01-01T01:02:03.123',
        'thread_identifier': 1234}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    # Check entry with recorded time with a year and milliseconds precision
    # with a time zone offset.
    expected_event_values = {
        'component_tag': 'App',
        'data_type': 'android:logcat',
        'message': 'test of threadtime w/ UTC and year',
        'pid': 1885,
        'priority': 'E',
        'recorded_time': '2022-01-02T01:20:03.171+00:00',
        'thread_identifier': 3066,
        'user_identifier': None}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 2)
    self.CheckEventData(event_data, expected_event_values)

    # Check entry with recorded time with a year and microseconds precision
    # in local time.
    expected_event_values = {
        'component_tag': 'App',
        'data_type': 'android:logcat',
        'message': 'test of threadtime w/ year, uid, usec',
        'pid': 1885,
        'priority': 'E',
        'recorded_time': '2022-01-02T01:20:07.171997',
        'thread_identifier': 3066,
        'user_identifier': 1000}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 4)
    self.CheckEventData(event_data, expected_event_values)

    # Check entry with recorded time with a year and microseconds precision
    # with time a zone offset.
    expected_event_values = {
        'component_tag': 'AppTag',
        'data_type': 'android:logcat',
        'message': 'test of threadtime w/ year, zone, usec',
        'pid': 9346,
        'priority': 'E',
        'recorded_time': '2022-01-02T11:20:07.171997+10:00',
        'thread_identifier': 9347,
        'user_identifier': None}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 5)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the sessionize analysis plugin."""

import collections
import unittest

from plaso.analysis import sessionize
from plaso.containers import reports
from plaso.lib import definitions

from tests.analysis import test_lib


class SessionizeAnalysisPluginTest(test_lib.AnalysisPluginTestCase):
  """Tests the sessionize analysis plugin."""

  _TEST_EVENTS = [
      {'_parser_chain': 'test_parser',
       'data_type': 'test:event',
       'timestamp': '2015-05-01 00:00:00',
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN},
      {'_parser_chain': 'test_parser',
       'data_type': 'test:event',
       'timestamp': '2015-05-01 00:09:00',
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN},
      {'_parser_chain': 'test_parser',
       'data_type': 'test:event',
       'timestamp': '2015-05-01 00:18:00',
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN},
      {'_parser_chain': 'test_parser',
       'data_type': 'test:event',
       'timestamp': '2015-05-01 01:00:00',
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN},
      {'_parser_chain': 'test_parser',
       'data_type': 'test:event',
       'timestamp': '2015-05-01 01:09:00',
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN}]

  def testTagAndCompileReport(self):
    """Tests the Sessionize plugin."""
    plugin = sessionize.SessionizeAnalysisPlugin()
    plugin.SetMaximumPause(10)

    storage_writer = self._AnalyzeEvents(self._TEST_EVENTS, plugin)

    number_of_reports = storage_writer.GetNumberOfAttributeContainers(
        'analysis_report')
    self.assertEqual(number_of_reports, 1)

    analysis_report = storage_writer.GetAttributeContainerByIndex(
        reports.AnalysisReport.CONTAINER_TYPE, 0)
    self.assertIsNotNone(analysis_report)

    self.assertEqual(analysis_report.plugin_name, 'sessionize')

    expected_analysis_counter = collections.Counter({
        'session_0': 3,
        'session_1': 2})
    self.assertEqual(
        analysis_report.analysis_counter, expected_analysis_counter)

    number_of_event_tags = storage_writer.GetNumberOfAttributeContainers(
        'event_tag')
    self.assertEqual(number_of_event_tags, 5)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the line-based JSON (JSON-L) log format parser."""

import unittest

from plaso.parsers import jsonl_parser
from plaso.parsers import jsonl_plugins  # pylint: disable=unused-import

from tests.parsers import test_lib


class JSONLParserTest(test_lib.ParserTestCase):
  """Tests for the line-based JSON (JSON-L) log format parser."""

  # pylint: disable=protected-access

  def testEnablePlugins(self):
    """Tests the EnablePlugins function."""
    parser = jsonl_parser.JSONLParser()

    number_of_plugins = len(parser._plugin_classes)

    parser.EnablePlugins([])
    self.assertEqual(len(parser._plugins_per_name), 0)

    parser.EnablePlugins(parser.ALL_PLUGINS)
    self.assertEqual(len(parser._plugins_per_name), number_of_plugins)

    parser.EnablePlugins(['gcp_log'])
    self.assertEqual(len(parser._plugins_per_name), 1)

  def testParse(self):
    """Tests the Parse function."""
    parser = jsonl_parser.JSONLParser()
    storage_writer = self._ParseFile(['gcp_logging.jsonl'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 9)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The status view CLI arguments helper."""

from plaso.cli import status_view
from plaso.cli import tools
from plaso.cli.helpers import interface
from plaso.cli.helpers import manager
from plaso.lib import errors


class StatusViewArgumentsHelper(interface.ArgumentsHelper):
  """Status view CLI arguments helper."""

  NAME = 'status_view'
  DESCRIPTION = 'Status view command line arguments.'

  _STATUS_VIEW_TYPES = ['file', 'linear', 'none', 'window']

  @classmethod
  def AddArguments(cls, argument_group):
    """Adds command line arguments to an argument group.

    This function takes an argument parser or an argument group object and adds
    to it all the command line arguments this helper supports.

    Args:
      argument_group (argparse._ArgumentGroup|argparse.ArgumentParser):
          argparse group.
    """
    argument_group.add_argument(
        '--status_view', '--status-view', dest='status_view_mode',
        choices=cls._STATUS_VIEW_TYPES, action='store', metavar='TYPE',
        default=status_view.StatusView.MODE_WINDOW, help=(
            'The processing status view mode: "file", "linear", "none" or '
            '"window".'))

    argument_group.add_argument(
        '--status_view_file', '--status-view-file', dest='status_view_file',
        action='store', metavar='PATH', default='status.info', help=(
            'The name of the status view file.'))

    argument_group.add_argument(
        '--status_view_interval', '--status-view-interval',
        dest='status_view_interval', action='store', type=float,
        metavar='SECONDS', default=0.5, help=(
            'Number of seconds to update the status view.'))

  @classmethod
  def ParseOptions(cls, options, configuration_object):
    """Parses and validates options.

    Args:
      options (argparse.Namespace): parser options.
      configuration_object (CLITool): object to be configured by the argument
          helper.

    Raises:
      BadConfigObject: when the configuration object is of the wrong type.
      BadConfigOption: when a configuration parameter fails validation.
    """
    if not isinstance(configuration_object, tools.CLITool):
      raise errors.BadConfigObject(
          'Configuration object is not an instance of CLITool')

    status_view_mode = cls._ParseStringOption(
        options, 'status_view_mode',
        default_value=status_view.StatusView.MODE_WINDOW)

    status_view_file = cls._ParseStringOption(
        options, 'status_view_file', default_value='status.info')

    status_view_interval = cls._ParseNumericOption(
        options, 'status_view_interval')

    if status_view_interval is None or status_view_interval <= 0.0:
      raise errors.BadConfigOption(
          'Invalid status view interval value must be larger than 0.0 seconds.')

    setattr(configuration_object, '_status_view_mode', status_view_mode)
    setattr(configuration_object, '_status_view_file', status_view_file)
    setattr(configuration_object, '_status_view_interval', status_view_interval)


manager.ArgumentHelperManager.RegisterHelper(StatusViewArgumentsHelper)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Google Drive Sync log log text parser plugin."""

import unittest

from dfvfs.helpers import fake_file_system_builder

from plaso.parsers import text_parser
from plaso.parsers.text_plugins import gdrive_synclog

from tests.parsers.text_plugins import test_lib


class GoogleDriveSyncLogTextPluginTest(test_lib.TextPluginTestCase):
  """Tests for the Google Drive Sync log text parser plugin."""

  def testCheckRequiredFormat(self):
    """Tests for the CheckRequiredFormat method."""
    plugin = gdrive_synclog.GoogleDriveSyncLogTextPlugin()

    file_system_builder = fake_file_system_builder.FakeFileSystemBuilder()
    file_system_builder.AddFile('/file.txt', (
        b'2018-01-24 18:25:08,454 -0800 INFO pid=2376 7780:MainThread      '
        b'logging_config.py:295 OS: Windows/6.1-SP1\n'))

    file_entry = file_system_builder.file_system.GetFileEntryByPath('/file.txt')

    parser_mediator = self._CreateParserMediator(None, file_entry=file_entry)

    file_object = file_entry.GetFileObject()
    text_reader = text_parser.EncodedTextReader(file_object)
    text_reader.ReadLines()

    result = plugin.CheckRequiredFormat(parser_mediator, text_reader)
    self.assertTrue(result)

  def testProcess(self):
    """Tests the Process function."""
    plugin = gdrive_synclog.GoogleDriveSyncLogTextPlugin()
    storage_writer = self._ParseTextFileWithPlugin([
        'gdrive_synclog', 'sync_log.log'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 2190)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'added_time': '2018-01-24T18:25:08.456-08:00',
        'data_type': 'google_drive_sync_log:entry',
        'level': 'INFO',
        'message': 'SSL: OpenSSL 1.0.2m  2 Nov 2017',
        'process_identifier': 2376,
        'source_code': 'logging_config.py:299',
        'thread': '7780:MainThread'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 2)
    self.CheckEventData(event_data, expected_event_values)

  def testProcessWithMacOSLog(self):
    """Tests the Process function with a MacOS Google Drive sync log."""
    plugin = gdrive_synclog.GoogleDriveSyncLogTextPlugin()
    storage_writer = self._ParseTextFileWithPlugin([
        'gdrive_synclog', 'sync_log-osx.log'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 2338)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'added_time': '2018-03-01T12:48:14.224-08:00',
        'data_type': 'google_drive_sync_log:entry',
        'level': 'INFO',
        'message': 'OS: Darwin/10.13.3',
        'process_identifier': 1730,
        'source_code': 'logging_config.pyo:295',
        'thread': '140736280556352:MainThread'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    # Test change in local system time from -0800 to UTC.
    # The switch occurs around line 215.
    expected_event_values = {
        'added_time': '2018-03-01T20:57:33.499+00:00',
        'data_type': 'google_drive_sync_log:entry',
        'level': 'INFO',
        'message': 'SSL: OpenSSL 1.0.2n  7 Dec 2017',
        'process_identifier': 2590,
        'source_code': 'logging_config.pyo:299',
        'thread': '140736280556352:MainThread'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 169)
    self.CheckEventData(event_data, expected_event_values)

    # Test with Unicode characters in filename.
    expected_event_values = {
        'added_time': '2018-03-05T03:09:15.806+00:00',
        'data_type': 'google_drive_sync_log:entry',
        'level': 'INFO',
        'message': (
            'Updating local entry local_id=LocalID(inode=870321, volume='
            '\'60228B87-A626-4F5C-873E-476615F863C6\'), filename=АБВГДЕ.gdoc, '
            'modified=1520218963, checksum=ab0618852c5d671d7b1b9191aef03bda, '
            'size=185, is_folder=False'),
        'process_identifier': 2608,
        'source_code': 'snapshot_sqlite.pyo:219',
        'thread': '123145558327296:Worker-1'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 1400)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Parser for Apple Spotlight store database files."""

from collections import abc as collections

import abc
import os
import zlib

from dfdatetime import cocoa_time as dfdatetime_cocoa_time
from dfdatetime import posix_time as dfdatetime_posix_time

from dtfabric import errors as dtfabric_errors
from dtfabric.runtime import data_maps as dtfabric_data_maps

import lz4.block

from plaso.containers import events
from plaso.lib import dtfabric_helper
from plaso.lib import errors
from plaso.lib import specification
from plaso.parsers import interface
from plaso.parsers import manager


class SpotlightStoreIndexValue(object):
  """Index value.

  Attributes:
    table_index (int): table index.
    values_list (list[str]): values list.
  """

  def __init__(self):
    """Initializes an index value."""
    super(SpotlightStoreIndexValue, self).__init__()
    self.table_index = None
    self.values_list = []


class SpotlightStoreMetadataItemEventData(events.EventData):
  """Apple Spotlight store database metadata item event data.

  Attributes:
    added_time (dfdatetime.DateTimeValues): date and time the item was added
        (kMDItemDateAdded).
    attribute_change_time (dfdatetime.DateTimeValues): date and time
        an attribute was last changed (kMDItemAttributeChangeDate).
    content_creation_time (dfdatetime.DateTimeValues): date and time
        the content was created (kMDItemContentCreationDate).
    content_modification_time (dfdatetime.DateTimeValues): date and time
        the content was last modified (kMDItemContentModificationDate).
    content_type (str): content type of the corresponding file (system) entry
        (kMDItemContentType).
    creation_time (dfdatetime.DateTimeValues): date and time the item was
        created (_kMDItemCreationDate).
    downloaded_time (dfdatetime.DateTimeValues): date and time the item was
        downloaded (kMDItemDownloadedDate).
    file_name (str): name of the corresponding file (system) entry
        (_kMDItemFileName).
    file_system_identifier (int): file system identifier, for example the
        catalog node identifier (CNID) on HFS.
    kind (str): item kind (kMDItemKind).
    modification_time (dfdatetime.DateTimeValues): date and time the item was
        last modified (_kMDItemContentChangeDate).
    parent_file_system_identifier (int): file system identifier of the parent.
    purchase_time  (dfdatetime.DateTimeValues): date and time the item was
        purchased in the AppStore (kMDItemAppStorePurchaseDate).
    snapshot_times (list[dfdatetime.DateTimeValues]): dates and times of
        the creation of backup snaphots (_kTimeMachineOldestSnapshot and
        _kTimeMachineNewestSnapshot).
    update_time (dfdatetime.DateTimeValues): date and time the item was last
        updated.
    used_times (list[dfdatetime.DateTimeValues]): dates and times when the item
        was used (kMDItemUsedDates and kMDItemLastUsedDate).
  """

  DATA_TYPE = 'spotlight:metadata_item'

  def __init__(self):
    """Initializes event data."""
    super(SpotlightStoreMetadataItemEventData, self).__init__(
        data_type=self.DATA_TYPE)
    self.added_time = None
    self.attribute_change_time = None
    self.content_creation_time = None
    self.content_modification_time = None
    self.content_type = None
    self.creation_time = None
    self.downloaded_time = None
    self.file_name = None
    self.file_system_identifier = None
    self.kind = None
    self.modification_time = None
    self.parent_file_system_identifier = None
    self.purchase_time  = None
    self.snapshot_times = None
    self.update_time = None
    self.used_times = None


class SpotlightStoreMetadataAttribute(object):
  """Metadata attribute.

  Attributes:
    key (str): key or name of the metadata attribute.
    property_type (int): metadata attribute property type.
    value (object): metadata attribute value.
    value_type (int): metadata attribute value type.
  """

  def __init__(self):
    """Initializes a metadata attribute."""
    super(SpotlightStoreMetadataAttribute, self).__init__()
    self.key = None
    self.property_type = None
    self.value = None
    self.value_type = None


class SpotlightStoreMetadataItem(object):
  """Metadata item.

  Attributes:
    attributes (dict[str, SpotlightStoreMetadataAttribute]): metadata
        attributes.
    data_size (int): size of the record data.
    flags (int): record flags.
    identifier (int): file (system) entry identifier.
    item_identifier (int): item identifier.
    last_update_time (int): last update time.
    parent_identifier (int): parent file (system) entry identifier.
  """

  def __init__(self):
    """Initializes a record."""
    super(SpotlightStoreMetadataItem, self).__init__()
    self.attributes = {}
    self.data_size = 0
    self.flags = 0
    self.identifier = 0
    self.item_identifier = 0
    self.last_update_time = 0
    self.parent_identifier = 0


class BaseSpotlightFile(dtfabric_helper.DtFabricHelper):
  """Shared functionality for Apple Spotlight files."""

  _DEFINITION_FILE = os.path.join(
      os.path.dirname(__file__), 'spotlight_storedb.yaml')

  def __init__(self):
    """Initializes a Apple Spotlight file."""
    super(BaseSpotlightFile, self).__init__()
    self._file_entry = None
    self._file_object = None

  def Close(self):
    """Closes an Apple Spotlight file.

    Raises:
      IOError: if the file is not opened.
      OSError: if the file is not opened.
    """
    if not self._file_object:
      raise IOError('File not opened')

    self._file_object = None
    self._file_entry = None

  def Open(self, file_entry):
    """Opens an Apple Spotlight file.

    Args:
      file_entry (dfvfs.FileEntry): a file entry.

    Raises:
      IOError: if the file is already opened.
      OSError: if the file is already opened.
    """
    if self._file_object:
      raise IOError('File already opened')

    self._file_entry = file_entry

    file_object = file_entry.GetFileObject()

    self.ReadFileObject(file_object)

    self._file_object = file_object

  @abc.abstractmethod
  def ReadFileObject(self, file_object):
    """Reads an Apple Spotlight file-like object.

    Args:
      file_object (file): file-like object.
    """


class SpotlightStreamsMapDataFile(BaseSpotlightFile):
  """Apple Spotlight database streams map data file (dbStr-#.map.data).

  Attributes:
    stream_values (list[bytes]): stream values.
  """

  def __init__(self, data_size, ranges):
    """Initializes a database streams map data file.

    Args:
      data_size (int): data size.
      ranges (list[tuple[int, int]]): offset and size pairs of the stream value
          data ranges.
    """
    super(SpotlightStreamsMapDataFile, self).__init__()
    self._data_size = data_size
    self._ranges = ranges
    self.stream_values = []

  def _ReadVariableSizeInteger(self, data):
    """Reads a variable size integer.

    Args:
      data (bytes): data.

    Returns:
      tuple[int, int]: integer value and number of bytes read.
    """
    byte_value = data[0]
    bytes_read = 1

    number_of_additional_bytes = 0
    for bitmask in (0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff):
      if byte_value & bitmask != bitmask:
        break
      number_of_additional_bytes += 1

    if number_of_additional_bytes > 4:
      byte_value = 0
    elif number_of_additional_bytes > 0:
      byte_value &= bitmask ^ 0xff

    integer_value = int(byte_value)
    while number_of_additional_bytes > 0:
      integer_value <<= 8

      integer_value += int(data[bytes_read])
      bytes_read += 1

      number_of_additional_bytes -= 1

    return integer_value, bytes_read

  def ReadFileObject(self, file_object):
    """Reads a database streams map data file-like object.

    Args:
      file_object (file): file-like object.

    Raises:
      ParseError: if the file cannot be read.
    """
    data = file_object.read(self._data_size)

    for value_range in self._ranges:
      value_offset, value_size = value_range

      stream_value = data[value_offset:value_offset + value_size]

      self.stream_values.append(stream_value)


class SpotlightStreamsMapHeaderFile(BaseSpotlightFile):
  """Apple Spotlight database streams map header file (dbStr-#.map.header).

  Attributes:
    data_size (int): data size.
    number_of_buckets (int): number of entries in the database streams map
        buckets file (dbStr-#.map.buckets).
    number_of_offsets (int): number of entries in the database streams map
        offsets file (dbStr-#.map.offsets).
  """

  def __init__(self):
    """Initializes a database streams map header file."""
    super(SpotlightStreamsMapHeaderFile, self).__init__()
    self.data_size = None
    self.number_of_buckets = None
    self.number_of_offsets = None

  def ReadFileObject(self, file_object):
    """Reads a database streams map header file-like object.

    Args:
      file_object (file): file-like object.

    Raises:
      ParseError: if the file cannot be read.
    """
    data_type_map = self._GetDataTypeMap(
        'spotlight_database_streams_map_header')

    streams_map_header, _ = self._ReadStructureFromFileObject(
        file_object, 0, data_type_map)

    self.data_size = streams_map_header.unknown4
    self.number_of_buckets = streams_map_header.unknown5
    self.number_of_offsets = streams_map_header.unknown6


class SpotlightStreamsMapOffsetsFile(BaseSpotlightFile):
  """Apple Spotlight database streams map offsets file (dbStr-#.map.offsets).

  Attributes:
    ranges (list[tuple[int, int]]): offset and size pairs of the stream value
        data ranges.
  """

  def __init__(self, data_size, number_of_entries):
    """Initializes a database streams map offsets file.

    Args:
      data_size (int): data size.
      number_of_entries (int): number of entries in the offsets file.
    """
    super(SpotlightStreamsMapOffsetsFile, self).__init__()
    self._data_size = data_size
    self._number_of_entries = number_of_entries
    self.ranges = []

  def ReadFileObject(self, file_object):
    """Reads a database streams map offsets file-like object.

    Args:
      file_object (file): file-like object.

    Raises:
      ParseError: if the file cannot be read.
    """
    data_size = self._number_of_entries * 4
    data = file_object.read(data_size)

    data_type_map = self._GetDataTypeMap('array_of_uint32le')

    context = dtfabric_data_maps.DataTypeMapContext(values={
        'number_of_elements': self._number_of_entries})

    try:
      offsets_array = data_type_map.MapByteStream(data, context=context)

    except dtfabric_errors.MappingError as exception:
      raise errors.ParseError(
          f'Unable to parse array of 32-bit offsets with error: {exception!s}')

    index = 0
    last_offset = 0

    for index, offset in enumerate(offsets_array):
      if index == 0:
        last_offset = offsets_array[0]
        continue

      range_size = offset - last_offset

      self.ranges.append((last_offset, range_size))

      last_offset = offset

    if last_offset:
      range_size = self._data_size - last_offset

      self.ranges.append((last_offset, range_size))


class SpotlightStoreDatabaseParser(
    interface.FileEntryParser, dtfabric_helper.DtFabricHelper):
  """Parser for Apple Spotlight store database (store.db) files."""

  NAME = 'spotlight_storedb'
  DATA_FORMAT = 'Apple Spotlight store database (store.db) file'

  _DEFINITION_FILE = os.path.join(
      os.path.dirname(__file__), 'spotlight_storedb.yaml')

  def __init__(self):
    """Initializes an Apple Spotlight store database parser."""
    super(SpotlightStoreDatabaseParser, self).__init__()
    self._map_values = []
    self._metadata_lists = {}
    self._metadata_localized_strings = {}
    self._metadata_types = {}
    self._metadata_values = {}

  def _DecompressLZ4Block(
      self, file_offset, compressed_data, previous_uncompressed_data):
    """Decompresses LZ4 compressed block.

    Args:
      file_offset (int): file offset.
      compressed_data (bytes): LZ4 compressed data.
      previous_uncompressed_data (bytes): uncompressed data of the previous
          (preceding) block.

    Returns:
      tuple[bytes, int]: uncompressed data and number of bytes read.

    Raises:
      ParseError: if the data cannot be decompressed.
    """
    data_type_map = self._GetDataTypeMap('spotlight_store_db_lz4_block_header')

    try:
      lz4_block_header = data_type_map.MapByteStream(compressed_data)
    except dtfabric_errors.MappingError as exception:
      raise errors.ParseError((
          f'Unable to map LZ4 block header at offset: 0x{file_offset:08x} '
          f'with error: {exception!s}'))

    if lz4_block_header.signature == b'bv41':
      end_of_data_offset = 12 + lz4_block_header.compressed_data_size

      uncompressed_data = lz4.block.decompress(
          compressed_data[12:end_of_data_offset],
          uncompressed_size=lz4_block_header.uncompressed_data_size,
          dict=previous_uncompressed_data)

    elif lz4_block_header.signature == b'bv4-':
      end_of_data_offset = 8 + lz4_block_header.uncompressed_data_size
      uncompressed_data = compressed_data[8:end_of_data_offset]

    else:
      raise errors.ParseError((
          f'Unsupported start of LZ4 block marker at offset: '
          f'0x{file_offset:08x}'))

    return uncompressed_data, end_of_data_offset

  def _DecompressLZ4PageData(self, compressed_page_data, file_offset):
    """Decompresses LZ4 compressed page data.

    Args:
      compressed_page_data (bytes): LZ4 compressed page data.
      file_offset (int): file offset.

    Returns:
      bytes: uncompressed page data.

    Raises:
      ParseError: if the page data cannot be decompressed.
    """
    compressed_data_offset = 0
    compressed_data_size = len(compressed_page_data)

    last_uncompressed_block = None

    uncompressed_blocks = []
    while compressed_data_offset < compressed_data_size:
      lz4_block_marker = compressed_page_data[
          compressed_data_offset:compressed_data_offset + 4]

      if lz4_block_marker == b'bv4$':
        break

      uncompressed_data, bytes_read = self._DecompressLZ4Block(
          file_offset, compressed_page_data[compressed_data_offset:],
          last_uncompressed_block)

      compressed_data_offset += bytes_read
      file_offset += bytes_read

      last_uncompressed_block = uncompressed_data

      uncompressed_blocks.append(uncompressed_data)

    if lz4_block_marker != b'bv4$':
      raise errors.ParseError(
          f'Unsupported end of LZ4 block marker at offset: 0x{file_offset:08x}')

    return b''.join(uncompressed_blocks)

  def _GetDateTimeMetadataItemValue(self, metadata_item, name):
    """Retrieves a date and time value from a metadata item.

    Args:
      metadata_item (SpotlightStoreMetadataItem): a metadata item.
      name (str): name of the attribute.

    Returns:
      dfdatetime.CocoaTime: a date and time value or None if not available.
    """
    value = self._GetMetadataItemValue(metadata_item, name)
    if not value:
      return None

    if isinstance(value, collections.Sequence):
      return dfdatetime_cocoa_time.CocoaTime(timestamp=value[0])

    return dfdatetime_cocoa_time.CocoaTime(timestamp=value)

  def _GetDateTimeMetadataItemValues(self, metadata_item, name):
    """Retrieves a date and time value from a metadata item.

    Args:
      metadata_item (SpotlightStoreMetadataItem): a metadata item.
      name (str): name of the attribute.

    Returns:
      list[dfdatetime.CocoaTime]: date and time values or None if not available.
    """
    value = self._GetMetadataItemValue(metadata_item, name)
    if not value:
      return None

    return [dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)
            for timestamp in value]

  def _GetMetadataItemValue(self, metadata_item, name):
    """Retrieves a value from a metadata item.

    Args:
      metadata_item (SpotlightStoreMetadataItem): a metadata item.
      name (str): name of the attribute.

    Returns:
      object: a time value or None if not available.
    """
    metadata_attribute = metadata_item.attributes.get(name, None)
    if not metadata_attribute:
      return None

    return metadata_attribute.value

  def _ParseMetadataItem(self, parser_mediator, metadata_item):
    """Parses an Apple Spotlight store metadata item.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      metadata_item (SpotlightStoreMetadataItem): a metadata item.
    """
    event_data = SpotlightStoreMetadataItemEventData()

    # TODO: for identifier 1 extract and process kMDStoreProperties plist

    # Identifier 1 is used for volume metadata.
    if metadata_item.identifier > 1:
      event_data.file_system_identifier = metadata_item.identifier
      event_data.parent_file_system_identifier = metadata_item.parent_identifier

    snapshot_times = []

    date_time = self._GetDateTimeMetadataItemValue(
        metadata_item, '_kTimeMachineOldestSnapshot')
    if date_time:
      snapshot_times.append(date_time)

    date_time = self._GetDateTimeMetadataItemValue(
        metadata_item, '_kTimeMachineNewestSnapshot')
    if date_time:
      snapshot_times.append(date_time)

    used_times = self._GetDateTimeMetadataItemValues(
        metadata_item, 'kMDItemUsedDates') or []

    date_time = self._GetDateTimeMetadataItemValue(
        metadata_item, 'kMDItemLastUsedDate')
    if date_time:
      used_times.append(date_time)

    # TODO: add more attribute values.
    event_data.added_time = self._GetDateTimeMetadataItemValue(
        metadata_item, 'kMDItemDateAdded')
    event_data.attribute_change_time = self._GetDateTimeMetadataItemValue(
        metadata_item, 'kMDItemAttributeChangeDate')
    event_data.content_creation_time = self._GetDateTimeMetadataItemValue(
        metadata_item, 'kMDItemContentCreationDate')
    event_data.content_modification_time = self._GetDateTimeMetadataItemValue(
        metadata_item, 'kMDItemContentModificationDate')
    event_data.content_type = self._GetMetadataItemValue(
        metadata_item, 'kMDItemContentType')
    event_data.creation_time = self._GetDateTimeMetadataItemValue(
        metadata_item, '_kMDItemCreationDate')
    event_data.downloaded_time = self._GetDateTimeMetadataItemValue(
        metadata_item, 'kMDItemDownloadedDate')
    event_data.file_name = self._GetMetadataItemValue(
        metadata_item, '_kMDItemFileName')
    event_data.kind = self._GetMetadataItemValue(
        metadata_item, 'kMDItemKind')
    event_data.modification_time = self._GetDateTimeMetadataItemValue(
        metadata_item, '_kMDItemContentChangeDate')
    event_data.purchase_time = self._GetDateTimeMetadataItemValue(
        metadata_item, 'kMDItemAppStorePurchaseDate')
    event_data.snapshot_times = snapshot_times or None
    event_data.used_times = used_times or None

    if metadata_item.last_update_time:
      event_data.update_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
          timestamp=metadata_item.last_update_time)

    parser_mediator.ProduceEventData(event_data)

  def _ParseRecord(self, parser_mediator, page_data, page_data_offset):
    """Parses a record for its metadata item.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      page_data (bytes): page data.
      page_data_offset (int): offset of the page value relative to the start
          of the page data.

    Returns:
      int: number of bytes read.

    Raises:
      ParseError: if the record cannot be read.
    """
    metadata_item, bytes_read = self._ReadRecordHeader(
        page_data[page_data_offset:], page_data_offset)

    page_data_offset += bytes_read
    record_data_offset = bytes_read

    metadata_type_index = 0

    try:
      while record_data_offset < metadata_item.data_size:
        relative_metadata_type_index, bytes_read = (
            self._ReadVariableSizeInteger(page_data[page_data_offset:]))

        page_data_offset += bytes_read
        record_data_offset += bytes_read

        metadata_type_index += relative_metadata_type_index

        metadata_type = self._metadata_types.get(metadata_type_index, None)
        metadata_attribute, bytes_read = self._ReadMetadataAttribute(
            metadata_type, page_data[page_data_offset:])

        page_data_offset += bytes_read
        record_data_offset += bytes_read

        metadata_item.attributes[metadata_attribute.key] = metadata_attribute

      self._ParseMetadataItem(parser_mediator, metadata_item)

    except errors.ParseError as exception:
      parser_mediator.ProduceExtractionWarning(
          'unable to read metadata item: 0x{0:2x} with error: {1!s}'.format(
              metadata_item.identifier, exception))

    return 4 + metadata_item.data_size

  def _ParseRecordPageValues(self, parser_mediator, page_data):
    """Parses the record page values.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      page_data (bytes): page data.

    Raises:
      ParseError: if the property page values cannot be read.
    """
    page_data_offset = 0
    page_data_size = len(page_data)

    while page_data_offset < page_data_size:
      bytes_read = self._ParseRecord(
          parser_mediator, page_data, page_data_offset)

      page_data_offset += bytes_read

  def _ReadFileHeader(self, file_object):
    """Reads the file header.

    Args:
      file_object (file): file-like object.

    Returns:
      spotlight_store_db_file_header: file header.

    Raises:
      ParseError: if the file header cannot be read.
    """
    data_type_map = self._GetDataTypeMap('spotlight_store_db_file_header')

    file_header, _ = self._ReadStructureFromFileObject(
        file_object, 0, data_type_map)

    return file_header

  def _ReadIndexPageValues(self, page_header, page_data, property_table):
    """Reads the index page values.

    Args:
      page_header (spotlight_store_db_property_page_header): page header.
      page_data (bytes): page data.
      property_table (dict[int, SpotlightStoreIndexValue]): property table in
          which to store the property page values.

    Raises:
      ParseError: if the property page values cannot be read.
    """
    data_type_map = self._GetDataTypeMap('spotlight_store_db_property_value81')
    index_values_data_type_map = self._GetDataTypeMap(
        'spotlight_store_db_index_values')

    page_data_offset = 12
    page_data_size = page_header.used_page_size - 20
    page_value_index = 0

    while page_data_offset < page_data_size:
      try:
        property_value = data_type_map.MapByteStream(
            page_data[page_data_offset:])
      except dtfabric_errors.MappingError as exception:
        raise errors.ParseError((
            'Unable to map property value data at offset: 0x{0:08x} with '
            'error: {1!s}').format(page_data_offset, exception))

      page_value_size = 4

      index_size, bytes_read = self._ReadVariableSizeInteger(
          page_data[page_data_offset + page_value_size:])

      _, padding_size = divmod(index_size, 4)

      page_value_size += bytes_read + padding_size
      index_size -= padding_size

      context = dtfabric_data_maps.DataTypeMapContext(values={
          'index_size': index_size})

      try:
        index_values = index_values_data_type_map.MapByteStream(
            page_data[page_data_offset + page_value_size:], context=context)

      except dtfabric_errors.MappingError as exception:
        raise errors.ParseError((
            'Unable to parse index data at offset: 0x{0:08x} with error: '
            '{1:s}').format(page_data_offset + page_value_size, exception))

      page_value_size += index_size

      values_list = []
      for metadata_value_index in index_values:
        metadata_value = self._metadata_values.get(metadata_value_index, None)
        value_string = getattr(metadata_value, 'value_name', '')
        values_list.append(value_string)

      index_value = SpotlightStoreIndexValue()
      index_value.table_index = property_value.table_index
      index_value.values_list = values_list

      property_table[index_value.table_index] = index_value

      page_data_offset += page_value_size
      page_value_index += 1

  def _ReadIndexStreamsMap(
      self, parent_file_entry, streams_map_number, property_table):
    """Reads an index streams map.

    Args:
      parent_file_entry (dfvfs.FileEntry): parent of the Spotlight store.db
          file entry.
      streams_map_number (int): number of the streams map.
      property_table (dict[int, SpotlightStoreIndexValue]): property table in
          which to store the index values.

    Raises:
      ParseError: if the index streams map cannot be read.
    """
    stream_values = self._ReadStreamsMap(parent_file_entry, streams_map_number)

    index_values_data_type_map = self._GetDataTypeMap(
        'spotlight_store_db_index_values')

    for index, stream_value in enumerate(stream_values):
      if index == 0:
        continue

      _, data_offset = self._ReadVariableSizeInteger(stream_value)

      index_size, bytes_read = self._ReadVariableSizeInteger(
          stream_value[data_offset:])

      data_offset += bytes_read

      _, padding_size = divmod(index_size, 4)

      index_size -= padding_size

      context = dtfabric_data_maps.DataTypeMapContext(values={
          'index_size': index_size})

      try:
        index_values = index_values_data_type_map.MapByteStream(
            stream_value[data_offset + padding_size:], context=context)

      except dtfabric_errors.MappingError as exception:
        raise errors.ParseError((
            f'Unable to map stream value: {index:d} data with error: '
            f'{exception!s}'))

      values_list = []
      for metadata_value_index in index_values:
        metadata_value = self._metadata_values.get(metadata_value_index, None)
        value_string = getattr(metadata_value, 'value_name', '')
        values_list.append(value_string)

      index_value = SpotlightStoreIndexValue()
      index_value.table_index = index
      index_value.values_list = values_list

      property_table[index] = index_value

  def _ReadMapPage(self, file_object, file_offset):
    """Reads a map page.

    Args:
      file_object (file): file-like object.
      file_offset (int): file offset.

    Returns:
      spotlight_store_db_map_page_header: page header.

    Raises:
      ParseError: if the map page cannot be read.
    """
    data_type_map = self._GetDataTypeMap('spotlight_store_db_map_page_header')

    page_header, page_header_size = self._ReadStructureFromFileObject(
        file_object, file_offset, data_type_map)

    file_offset += page_header_size

    self._ReadMapPageValues(page_header, file_object, file_offset)

    return page_header

  def _ReadMapPages(self, file_object, map_offset, map_size):
    """Reads the map pages.

    Args:
      file_object (file): file-like object.
      map_offset (int): map offset.
      map_size (int): map size.

    Raises:
      ParseError: if the map pages cannot be read.
    """
    map_end_offset = map_offset + map_size

    while map_offset < map_end_offset:
      page_header = self._ReadMapPage(file_object, map_offset)

      map_offset += page_header.page_size

  def _ReadMapPageValues(self, page_header, file_object, file_offset):
    """Reads the map page values.

    Args:
      page_header (spotlight_store_db_map_page_header): page header.
      file_object (file): file-like object.
      file_offset (int): file offset.

    Raises:
      ParseError: if the map page values cannot be read.
    """
    data_type_map = self._GetDataTypeMap('spotlight_store_db_map_page_value')

    for _ in range(page_header.number_of_map_values):
      map_value, map_value_size = self._ReadStructureFromFileObject(
          file_object, file_offset, data_type_map)

      self._map_values.append(map_value)

      file_offset += map_value_size

  def _ReadMetadataAttribute(self, metadata_type, data):
    """Reads a metadata attribute.

    Args:
      metadata_type (spotlight_store_db_property_value11): metadata type
          property value.
      data (bytes): data.

    Returns:
      tuple[SpotlightStoreMetadataAttribute, int]: metadata attribute and
          number of bytes read.
    """
    value_type = getattr(metadata_type, 'value_type', None)
    if value_type is None:
      return None, 0

    key_name = getattr(metadata_type, 'key_name', None)
    property_type = getattr(metadata_type, 'property_type', None)

    if key_name == 'kMDStoreAccumulatedSizes':
      bytes_read = len(data)
      value = data

    elif value_type in (0x00, 0x02, 0x06):
      value, bytes_read = self._ReadVariableSizeInteger(data)

    elif value_type == 0x07:
      value, bytes_read = self._ReadMetadataAttributeVariableSizeIntegerValue(
          property_type, data)

    elif value_type == 0x08:
      value, bytes_read = self._ReadMetadataAttributeByteValue(
          property_type, data)

    elif value_type == 0x09:
      value, bytes_read = self._ReadMetadataAttributeFloat32Value(
          property_type, data)

    elif value_type in (0x0a, 0x0c):
      value, bytes_read = self._ReadMetadataAttributeFloat64Value(
          property_type, data)

    elif value_type == 0x0b:
      value, bytes_read = self._ReadMetadataAttributeStringValue(
          property_type, data)

    elif value_type == 0x0e:
      data_size, bytes_read = self._ReadVariableSizeInteger(data)

      value = data[bytes_read:bytes_read + data_size]
      bytes_read += data_size

      # TODO: decode binary data e.g. UUID

    elif value_type == 0x0f:
      value, bytes_read = self._ReadMetadataAttributeReferenceValue(
          property_type, data)

    else:
      # TODO: value type 0x01, 0x03, 0x04, 0x05, 0x0d
      value, bytes_read = None, 0

    metadata_attribute = SpotlightStoreMetadataAttribute()
    metadata_attribute.key = getattr(metadata_type, 'key_name', None)
    metadata_attribute.property_type = property_type
    metadata_attribute.value = value
    metadata_attribute.value_type = value_type

    return metadata_attribute, bytes_read

  def _ReadMetadataAttributeByteValue(self, property_type, data):
    """Reads a metadata attribute byte value.

    Args:
      property_type (int): metadata attribute property type.
      data (bytes): data.

    Returns:
      tuple[object, int]: value and number of bytes read.

    Raises:
      ParseError: if the metadata attribute byte value cannot be read.
    """
    if property_type & 0x02:
      data_size, bytes_read = self._ReadVariableSizeInteger(data)
    else:
      data_size, bytes_read = 1, 0

    data_type_map = self._GetDataTypeMap('array_of_byte')

    context = dtfabric_data_maps.DataTypeMapContext(values={
        'elements_data_size': data_size})

    try:
      array_of_values = data_type_map.MapByteStream(
          data[bytes_read:bytes_read + data_size], context=context)

    except dtfabric_errors.MappingError as exception:
      raise errors.ParseError(
          'Unable to parse array of byte values with error: {0!s}'.format(
              exception))

    if bytes_read == 0:
      value = array_of_values[0]
    else:
      value = array_of_values

    bytes_read += data_size

    return value, bytes_read

  def _ReadMetadataAttributeFloat32Value(self, property_type, data):
    """Reads a metadata attribute 32-bit floating-point value.

    Args:
      property_type (int): metadata attribute property type.
      data (bytes): data.

    Returns:
      tuple[object, int]: value and number of bytes read.

    Raises:
      ParseError: if the metadata attribute 32-bit floating-point value cannot
          be read.
    """
    if property_type & 0x02 == 0x00:
      data_size, bytes_read = 4, 0
    else:
      data_size, bytes_read = self._ReadVariableSizeInteger(data)

    data_type_map = self._GetDataTypeMap('array_of_float32')

    context = dtfabric_data_maps.DataTypeMapContext(values={
        'elements_data_size': data_size})

    try:
      array_of_values = data_type_map.MapByteStream(
          data[bytes_read:bytes_read + data_size], context=context)

    except dtfabric_errors.MappingError as exception:
      raise errors.ParseError((
          'Unable to parse array of 32-bit floating-point values with error: '
          '{0!s}').format(exception))

    if bytes_read == 0:
      value = array_of_values[0]
    else:
      value = array_of_values

    bytes_read += data_size

    return value, bytes_read

  def _ReadMetadataAttributeFloat64Value(self, property_type, data):
    """Reads a metadata attribute 64-bit floating-point value.

    Args:
      property_type (int): metadata attribute property type.
      data (bytes): data.

    Returns:
      tuple[object, int]: value and number of bytes read.

    Raises:
      ParseError: if the metadata attribute 64-bit floating-point value cannot
          be read.
    """
    if property_type & 0x02 == 0x00:
      data_size, bytes_read = 8, 0
    else:
      data_size, bytes_read = self._ReadVariableSizeInteger(data)

    data_type_map = self._GetDataTypeMap('array_of_float64')

    context = dtfabric_data_maps.DataTypeMapContext(values={
        'elements_data_size': data_size})

    try:
      array_of_values = data_type_map.MapByteStream(
          data[bytes_read:bytes_read + data_size], context=context)

    except dtfabric_errors.MappingError as exception:
      raise errors.ParseError((
          'Unable to parse array of 64-bit floating-point values with error: '
          '{0!s}').format(exception))

    if bytes_read == 0:
      value = array_of_values[0]
    else:
      value = array_of_values

    bytes_read += data_size

    return value, bytes_read

  def _ReadMetadataAttributePageValues(
      self, page_header, page_data, property_table):
    """Reads the metadata atribute page values.

    Args:
      page_header (spotlight_store_db_property_page_header): page header.
      page_data (bytes): page data.
      property_table (dict[int, object]): property table in which to store the
          property page values.

    Raises:
      ParseError: if the property page values cannot be read.
    """
    if page_header.property_table_type == 0x00000011:
      data_type_map = self._GetDataTypeMap(
          'spotlight_store_db_property_value11')

    elif page_header.property_table_type == 0x00000021:
      data_type_map = self._GetDataTypeMap(
          'spotlight_store_db_property_value21')

    else:
      data_type_map = None

    page_data_offset = 12
    page_data_size = page_header.used_page_size - 20
    page_value_index = 0

    while page_data_offset < page_data_size:
      context = dtfabric_data_maps.DataTypeMapContext()

      try:
        property_value = data_type_map.MapByteStream(
            page_data[page_data_offset:], context=context)
      except dtfabric_errors.MappingError as exception:
        raise errors.ParseError((
            'Unable to map property value data at offset: 0x{0:08x} with '
            'error: {1!s}').format(page_data_offset, exception))

      property_table[property_value.table_index] = property_value

      page_data_offset += context.byte_size
      page_value_index += 1

  def _ReadMetadataAttributeReferenceValue(self, property_type, data):
    """Reads a metadata attribute reference value.

    Args:
      property_type (int): metadata attribute property type.
      data (bytes): data.

    Returns:
      tuple[object, int]: value and number of bytes read.

    Raises:
      ParseError: if the metadata attribute reference value cannot be read.
    """
    table_index, bytes_read = self._ReadVariableSizeInteger(data)

    if property_type & 0x03 == 0x03:
      metadata_localized_strings = self._metadata_localized_strings.get(
          table_index, None)
      value_list = getattr(metadata_localized_strings, 'values_list', [])

      value = '(null)'
      if value_list:
        value = value_list[0]
        if '\x16\x02' in value:
          value = value.split('\x16\x02')[0]

    elif property_type & 0x03 == 0x02:
      metadata_list = self._metadata_lists.get(table_index, None)
      value = getattr(metadata_list, 'values_list', [])

    else:
      metadata_value = self._metadata_values.get(table_index, None)
      value = getattr(metadata_value, 'value_name', '(null)')

    return value, bytes_read

  def _ReadMetadataAttributeStringValue(self, property_type, data):
    """Reads a metadata attribute string value.

    Args:
      property_type (int): metadata attribute property type.
      data (bytes): data.

    Returns:
      tuple[object, int]: value and number of bytes read.

    Raises:
      ParseError: if the metadata attribute string value cannot be read.
    """
    data_size, bytes_read = self._ReadVariableSizeInteger(data)

    data_type_map = self._GetDataTypeMap('array_of_cstring')

    context = dtfabric_data_maps.DataTypeMapContext(values={
        'elements_data_size': data_size})

    try:
      array_of_values = data_type_map.MapByteStream(
          data[bytes_read:bytes_read + data_size], context=context)

    except dtfabric_errors.MappingError as exception:
      raise errors.ParseError(
          'Unable to parse array of string values with error: {0!s}'.format(
              exception))

    if property_type & 0x03 == 0x03:
      value = array_of_values[0]
      if '\x16\x02' in value:
        value = value.split('\x16\x02')[0]
    elif property_type & 0x03 == 0x02:
      value = array_of_values
    else:
      value = array_of_values[0]

    bytes_read += data_size

    return value, bytes_read

  def _ReadMetadataAttributeVariableSizeIntegerValue(self, property_type, data):
    """Reads a metadata attribute variable size integer value.

    Args:
      property_type (int): metadata attribute property type.
      data (bytes): data.

    Returns:
      tuple[object, int]: value and number of bytes read.
    """
    if property_type & 0x02 == 0x00:
      return self._ReadVariableSizeInteger(data)

    data_size, bytes_read = self._ReadVariableSizeInteger(data)

    array_of_values = []

    data_offset = 0
    while data_offset < data_size:
      integer_value, integer_value_size = self._ReadVariableSizeInteger(
          data[data_offset:data_size])

      data_offset += integer_value_size

      array_of_values.append(integer_value)

    bytes_read += data_size

    return array_of_values, bytes_read

  def _ReadMetadataAttributeStreamsMap(
      self, parent_file_entry, streams_map_number, property_table):
    """Reads a metadata attribute streams map.

    Args:
      parent_file_entry (dfvfs.FileEntry): parent of the Spotlight store.db
          file entry.
      streams_map_number (int): number of the streams map.
      property_table (dict[int, object]): property table in which to store the
          metadata attribute values.

    Raises:
      ParseError: if the metadata attribute streams map cannot be read.
    """
    stream_values = self._ReadStreamsMap(parent_file_entry, streams_map_number)

    if streams_map_number == 1:
      data_type_map = self._GetDataTypeMap('spotlight_metadata_attribute_type')
    elif streams_map_number == 2:
      data_type_map = self._GetDataTypeMap('spotlight_metadata_attribute_value')
    else:
      data_type_map = None

    for index, stream_value in enumerate(stream_values):
      if index == 0:
        continue

      data_size, data_offset = self._ReadVariableSizeInteger(stream_value)

      if data_offset + data_size != len(stream_value):
        # Stream values where the data size does not match appear to contain
        # remnant data.
        continue

      try:
        property_value = data_type_map.MapByteStream(stream_value[data_offset:])
      except dtfabric_errors.MappingError as exception:
        raise errors.ParseError((
            f'Unable to map stream value: {index:d} data with error: '
            f'{exception!s}'))

      property_value.table_index = index

      property_table[index] = property_value

  def _ReadPropertyPage(self, file_object, file_offset, property_table):
    """Reads a property page.

    Args:
      file_object (file): file-like object.
      file_offset (int): file offset.
      property_table (dict[int, object]): property table in which to store the
          property page values.

    Returns:
      tuple[spotlight_store_db_property_page_header, int]: page header and next
          property page block number.

    Raises:
      ParseError: if the property page cannot be read.
    """
    page_header, bytes_read = self._ReadPropertyPageHeader(
        file_object, file_offset)

    if page_header.property_table_type not in (
        0x00000011, 0x00000021, 0x00000041, 0x00000081):
      raise errors.ParseError(
          'Unsupported property table type: 0x{0:08x}'.format(
              page_header.property_table_type))

    page_data = file_object.read(page_header.page_size - bytes_read)

    data_type_map = self._GetDataTypeMap(
        'spotlight_store_db_property_values_header')

    file_offset += bytes_read

    page_values_header = self._ReadStructureFromByteStream(
        page_data, file_offset, data_type_map)

    if page_header.property_table_type in (0x00000011, 0x00000021):
      self._ReadMetadataAttributePageValues(
          page_header, page_data, property_table)

    elif page_header.property_table_type == 0x00000081:
      self._ReadIndexPageValues(page_header, page_data, property_table)

    return page_header, page_values_header.next_block_number

  def _ReadPropertyPageHeader(self, file_object, file_offset):
    """Reads a property page header.

    Args:
      file_object (file): file-like object.
      file_offset (int): file offset.

    Returns:
      tuple[spotlight_store_db_property_page_header, int]: page header and next
          property page block number.

    Raises:
      ParseError: if the property page header cannot be read.
    """
    data_type_map = self._GetDataTypeMap(
        'spotlight_store_db_property_page_header')

    return self._ReadStructureFromFileObject(
        file_object, file_offset, data_type_map)

  def _ReadPropertyPages(self, file_object, block_number, property_table):
    """Reads the property pages.

    Args:
      file_object (file): file-like object.
      block_number (int): block number.
      property_table (dict[int, object]): property table in which to store the
          property page values.

    Raises:
      ParseError: if the property pages cannot be read.
    """
    file_offset = block_number * 0x1000
    while file_offset != 0:
      _, next_block_number = self._ReadPropertyPage(
          file_object, file_offset, property_table)

      file_offset = next_block_number * 0x1000

  def _ReadRecordHeader(self, data, page_data_offset):
    """Reads a record header.

    Args:
      data (bytes): data.
      page_data_offset (int): offset of the page value relative to the start
          of the page data.

    Returns:
      tuple[SpotlightStoreMetadataItem, int]: metadata item and number of
          bytes read.

    Raises:
      ParseError: if the record page cannot be read.
    """
    data_type_map = self._GetDataTypeMap('spotlight_store_db_record')

    context = dtfabric_data_maps.DataTypeMapContext()

    try:
      record = data_type_map.MapByteStream(data, context=context)
    except dtfabric_errors.MappingError as exception:
      raise errors.ParseError((
          'Unable to map record at offset: 0x{0:08x} with error: '
          '{1!s}').format(page_data_offset, exception))

    data_offset = context.byte_size

    identifier, bytes_read = self._ReadVariableSizeInteger(data[data_offset:])

    data_offset += bytes_read

    flags = data[data_offset]

    data_offset += 1

    value_names = ['item_identifier', 'parent_identifier', 'last_update_time']
    values, bytes_read = self._ReadVariableSizeIntegers(
        data[data_offset:], value_names)

    data_offset += bytes_read

    metadata_item = SpotlightStoreMetadataItem()
    metadata_item.data_size = record.data_size
    metadata_item.flags = flags
    metadata_item.identifier = identifier
    metadata_item.item_identifier = values.get('item_identifier')
    metadata_item.last_update_time = values.get('last_update_time')
    metadata_item.parent_identifier = values.get('parent_identifier')

    return metadata_item, data_offset

  def _ReadRecordPage(self, file_object, file_offset):
    """Reads a record page.

    Args:
      file_object (file): file-like object.
      file_offset (int): file offset.

    Returns:
      tuple[spotlight_store_db_property_page_header, bytes]: page header and
          page data.

    Raises:
      ParseError: if the property page cannot be read.
    """
    page_header, bytes_read = self._ReadPropertyPageHeader(
        file_object, file_offset)

    if page_header.property_table_type not in (
        0x00000009, 0x00001009, 0x00005009):
      raise errors.ParseError(
          'Unsupported property table type: 0x{0:08x}'.format(
              page_header.property_table_type))

    page_data = file_object.read(page_header.page_size - bytes_read)

    file_offset += bytes_read

    if page_header.uncompressed_page_size > 0:
      compressed_page_data = page_data

      if (page_header.property_table_type & 0x00001000 and
          compressed_page_data[0:4] in (b'bv41', b'bv4-')):
        page_data = self._DecompressLZ4PageData(
            compressed_page_data, file_offset)

      elif compressed_page_data[0] == 0x78:
        page_data = zlib.decompress(compressed_page_data)

      else:
        raise errors.ParseError('Unsupported compression type')

    return page_header, page_data

  def _ReadStreamsMap(self, parent_file_entry, streams_map_number):
    """Reads a streams map.

    Args:
      parent_file_entry (dfvfs.FileEntry): parent of the Spotlight store.db
          file entry.
      streams_map_number (int): number of the streams map.

    Returns:
      list[bytes]: stream values.

    Raises:
      ParseError: if the streams map cannot be read.
    """
    header_file_entry = parent_file_entry.GetSubFileEntryByName(
        f'dbStr-{streams_map_number:d}.map.header')

    streams_map_header = SpotlightStreamsMapHeaderFile()
    streams_map_header.Open(header_file_entry)

    data_size = streams_map_header.data_size
    number_of_offsets = streams_map_header.number_of_offsets

    streams_map_header.Close()

    offsets_file_entry = parent_file_entry.GetSubFileEntryByName(
        f'dbStr-{streams_map_number:d}.map.offsets')

    streams_map_offsets = SpotlightStreamsMapOffsetsFile(
        data_size, number_of_offsets)
    streams_map_offsets.Open(offsets_file_entry)

    ranges = streams_map_offsets.ranges

    streams_map_offsets.Close()

    data_file_entry = parent_file_entry.GetSubFileEntryByName(
        f'dbStr-{streams_map_number:d}.map.data')

    streams_map_data = SpotlightStreamsMapDataFile(data_size, ranges)
    streams_map_data.Open(data_file_entry)

    stream_values = streams_map_data.stream_values

    streams_map_data.Close()

    return stream_values

  def _ReadVariableSizeInteger(self, data):
    """Reads a variable size integer.

    Args:
      data (bytes): data.

    Returns:
      tuple[int, int]: integer value and number of bytes read.
    """
    byte_value = data[0]
    bytes_read = 1

    number_of_additional_bytes = 0
    for bitmask in (0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff):
      if byte_value & bitmask != bitmask:
        break
      number_of_additional_bytes += 1

    if number_of_additional_bytes > 4:
      byte_value = 0
    elif number_of_additional_bytes > 0:
      byte_value &= bitmask ^ 0xff

    integer_value = int(byte_value)
    while number_of_additional_bytes > 0:
      integer_value <<= 8

      integer_value += int(data[bytes_read])
      bytes_read += 1

      number_of_additional_bytes -= 1

    return integer_value, bytes_read

  def _ReadVariableSizeIntegers(self, data, names):
    """Reads variable size integers.

    Args:
      data (bytes): data.
      names (list[str]): names to identify the integer values.

    Returns:
      tuple[dict[str, int], int]: integer values per name and number of bytes
          read.
    """
    values = {}

    data_offset = 0
    for name in names:
      integer_value, bytes_read = self._ReadVariableSizeInteger(
          data[data_offset:])

      data_offset += bytes_read

      values[name] = integer_value

    return values, data_offset

  @classmethod
  def GetFormatSpecification(cls):
    """Retrieves the format specification.

    Returns:
      FormatSpecification: format specification.
    """
    format_specification = specification.FormatSpecification(cls.NAME)
    format_specification.AddNewSignature(b'8tsd', offset=0)
    return format_specification

  def ParseFileEntry(self, parser_mediator, file_entry):
    """Parses an Apple Spotlight store database file entry.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_entry (dfvfs.FileEntry): a file entry to parse.

    Raises:
      WrongParser: when the file cannot be parsed.
    """
    parent_file_entry = file_entry.GetParentFileEntry()

    file_object = file_entry.GetFileObject()

    try:
      file_header = self._ReadFileHeader(file_object)
    except (ValueError, errors.ParseError):
      raise errors.WrongParser('Unable to parse file header.')

    try:
      self._map_values = []
      self._ReadMapPages(
          file_object, file_header.map_offset, file_header.map_size)

      self._metadata_types = {}
      if not file_header.metadata_types_block_number:
        self._ReadMetadataAttributeStreamsMap(
            parent_file_entry, 1, self._metadata_types)
      else:
        self._ReadPropertyPages(
            file_object, file_header.metadata_types_block_number,
            self._metadata_types)

      self._metadata_values = {}
      if not file_header.metadata_values_block_number:
        self._ReadMetadataAttributeStreamsMap(
            parent_file_entry, 2, self._metadata_values)
      else:
        self._ReadPropertyPages(
            file_object, file_header.metadata_values_block_number,
            self._metadata_values)

      # Note that the content of this property page is currently unknown.
      if not file_header.unknown_values41_block_number:
        self._ReadIndexStreamsMap(parent_file_entry, 3, {})
      else:
        self._ReadPropertyPages(
            file_object, file_header.unknown_values41_block_number, {})

      self._metadata_lists = {}
      if not file_header.metadata_lists_block_number:
        self._ReadIndexStreamsMap(parent_file_entry, 4, self._metadata_lists)
      else:
        self._ReadPropertyPages(
            file_object, file_header.metadata_lists_block_number,
            self._metadata_lists)

      self._metadata_localized_strings = {}
      if not file_header.metadata_localized_strings_block_number:
        self._ReadIndexStreamsMap(
            parent_file_entry, 5, self._metadata_localized_strings)
      else:
        self._ReadPropertyPages(
            file_object, file_header.metadata_localized_strings_block_number,
            self._metadata_localized_strings)

    except errors.ParseError as exception:
      parser_mediator.ProduceExtractionWarning(
          'unable to read store database with error: {0!s}'.format(
              exception))
      return

    for map_value in self._map_values:
      file_offset = map_value.block_number * 0x1000

      try:
        _, page_data = self._ReadRecordPage(file_object, file_offset)

        self._ParseRecordPageValues(parser_mediator, page_data)

      except errors.ParseError as exception:
        parser_mediator.ProduceExtractionWarning((
            'unable to read record page at offset: 0x{0:08x} with error: '
            '{1!s}').format(file_offset, exception))
        continue


manager.ParsersManager.RegisterParser(SpotlightStoreDatabaseParser)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for Twitter on iOS 8+ SQLite database plugin."""

import unittest

from plaso.parsers.sqlite_plugins import ios_twitter

from tests.parsers.sqlite_plugins import test_lib


class IOSTwitterTest(test_lib.SQLitePluginTestCase):
  """Tests for Twitter on iOS 8+ SQLite database plugin."""

  def testProcess(self):
    """Test the Process function."""
    plugin = ios_twitter.IOSTwitterPlugin()
    storage_writer = self._ParseDatabaseFileWithPlugin(
        ['twitter_ios.db'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 92)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    # Test a contact database entry.
    expected_event_values = {
        'creation_time': '2007-04-22T14:42:37+00:00',
        'data_type': 'ios:twitter:contact',
        'description': (
            'Breaking news alerts and updates from the BBC. For news, '
            'features, analysis follow @BBCWorld (international) or @BBCNews '
            '(UK). Latest sport news @BBCSport.'),
        'followers_count': 19466932,
        'following': 0,
        'following_count': 3,
        'location': 'London, UK',
        'modification_time': '2015-12-02T15:35:44+00:00',
        'name': 'BBC Breaking News',
        'profile_url': (
            'https://pbs.twimg.com/profile_images/460740982498013184/'
            'wIPwMwru_normal.png'),
        'screen_name': 'BBCBreaking',
        'url': 'http://www.bbc.co.uk/news'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    # Test a status database entry.
    expected_event_values = {
        'creation_time': '2014-09-11T11:46:16+00:00',
        'data_type': 'ios:twitter:status',
        'favorite_count': 3,
        'favorited': 0,
        'modification_time': '2015-12-02T15:39:37+00:00',
        'name': 'Heather Mahalik',
        'retweet_count': 2,
        'text': 'Never forget. http://t.co/L7bjWue1A2',
        'user_identifier': 475222380}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 25)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Base parser for line-based JSON (JSON-L) log formats."""

import json

from json import decoder as json_decoder

from dfvfs.helpers import text_file

from plaso.lib import errors
from plaso.parsers import interface
from plaso.parsers import manager


class JSONLParser(interface.FileObjectParser):
  """Base parser for line-based JSON (JSON-L) log formats."""

  NAME = 'jsonl'
  DATA_FORMAT = 'JSON-L log file'

  _ENCODING = 'utf-8'

  _MAXIMUM_LINE_LENGTH = 64 * 1024

  _plugin_classes = {}

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a line-based JSON (JSON-L) log file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_object (dfvfs.FileIO): a file-like object.

    Raises:
      WrongParser: when the file cannot be parsed.
    """
    encoding = self._ENCODING
    if not encoding:
      encoding = parser_mediator.GetCodePage()

    # Use strict encoding error handling in the verification step so that
    # a JSON-L parser does not generate extraction warning for encoding errors
    # of unsupported files.
    text_file_object = text_file.TextFile(file_object, encoding=encoding)

    try:
      line = text_file_object.readline(size=self._MAXIMUM_LINE_LENGTH)
    except UnicodeDecodeError:
      raise errors.WrongParser('Not a JSON-L file or encoding not supported.')

    if not line:
      raise errors.WrongParser('Not a JSON-L file.')

    line = line.strip()
    if not line or (line[0] != '{' and line[-1] != '}'):
      raise errors.WrongParser(
          'Not a JSON-L file, missing opening and closing braces.')

    try:
      json_dict = json.loads(line)
    except json_decoder.JSONDecodeError:
      raise errors.WrongParser('Not a JSON-L file, unable to decode JSON.')

    if not json_dict:
      raise errors.WrongParser('Not a JSON-L file, missing JSON.')

    for plugin_name, plugin in self._plugins_per_name.items():
      if parser_mediator.abort:
        break

      profiling_name = '/'.join([self.NAME, plugin.NAME])

      parser_mediator.SampleFormatCheckStartTiming(profiling_name)

      try:
        result = plugin.CheckRequiredFormat(json_dict)
      finally:
        parser_mediator.SampleFormatCheckStopTiming(profiling_name)

      if not result:
        continue

      parser_mediator.SampleStartTiming(profiling_name)

      try:
        plugin.UpdateChainAndProcess(
            parser_mediator, file_object=file_object)

      except Exception as exception:  # pylint: disable=broad-except
        parser_mediator.ProduceExtractionWarning((
            'plugin: {0:s} unable to parse JSON-L file with error: '
            '{1!s}').format(plugin_name, exception))

      finally:
        parser_mediator.SampleStopTiming(profiling_name)


manager.ParsersManager.RegisterParser(JSONLParser)



# -*- coding: utf-8 -*-
"""Engine related functions and classes for testing."""

from plaso.storage.fake import writer as fake_writer

from tests import test_lib as shared_test_lib


class EngineTestCase(shared_test_lib.BaseTestCase):
  """Engine test case."""

  def _CreateStorageWriter(self):
    """Creates a storage writer object.

    Returns:
      FakeStorageWriter: storage writer.
    """
    storage_writer = fake_writer.FakeStorageWriter()
    storage_writer.Open()
    return storage_writer



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the artifact definitions CLI arguments helper."""

import argparse
import unittest

from plaso.cli import tools
from plaso.cli.helpers import artifact_definitions
from plaso.lib import errors

from tests.cli import test_lib as cli_test_lib


class ArtifactDefinitionsArgumentsHelperTest(cli_test_lib.CLIToolTestCase):
  """Tests for the artifact definitions CLI arguments helper."""

  # pylint: disable=no-member,protected-access

  _EXPECTED_OUTPUT = """\
usage: cli_helper.py [--artifact_definitions PATH]
                     [--custom_artifact_definitions PATH]

Test argument parser.

{0:s}:
  --artifact_definitions PATH, --artifact-definitions PATH
                        Path to a directory or file containing artifact
                        definitions, which are .yaml files. Artifact
                        definitions can be used to describe and quickly
                        collect data of interest, such as specific files or
                        Windows Registry keys.
  --custom_artifact_definitions PATH, --custom-artifact-definitions PATH
                        Path to a directory or file containing custom artifact
                        definitions, which are .yaml files. Artifact
                        definitions can be used to describe and quickly
                        collect data of interest, such as specific files or
                        Windows Registry keys.
""".format(cli_test_lib.ARGPARSE_OPTIONS)

  def testAddArguments(self):
    """Tests the AddArguments function."""
    argument_parser = argparse.ArgumentParser(
        prog='cli_helper.py', description='Test argument parser.',
        add_help=False,
        formatter_class=cli_test_lib.SortedArgumentsHelpFormatter)

    artifact_definitions.ArtifactDefinitionsArgumentsHelper.AddArguments(
        argument_parser)

    output = self._RunArgparseFormatHelp(argument_parser)
    self.assertEqual(output, self._EXPECTED_OUTPUT)

  def testParseOptions(self):
    """Tests the ParseOptions function."""
    test_file_path = self._GetTestFilePath(['artifacts'])
    self._SkipIfPathNotExists(test_file_path)

    test_tool = tools.CLITool()

    options = cli_test_lib.TestOptions()
    options.artifact_definitions_path = test_file_path

    artifact_definitions.ArtifactDefinitionsArgumentsHelper.ParseOptions(
        options, test_tool)
    self.assertIsNotNone(test_tool._artifact_definitions_path)

    with self.assertRaises(errors.BadConfigObject):
      artifact_definitions.ArtifactDefinitionsArgumentsHelper.ParseOptions(
          options, None)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Text parser plugin for vsftpd log files."""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class VsftpdLogEventData(events.EventData):
  """Vsftpd log event data.

  Attributes:
    added_time (dfdatetime.DateTimeValues): date and time the log entry
        was added.
    text (str): vsftpd log message.
  """

  DATA_TYPE = 'vsftpd:log'

  def __init__(self):
    """Initializes event data."""
    super(VsftpdLogEventData, self).__init__(data_type=self.DATA_TYPE)
    self.added_time = None
    self.text = None


class VsftpdLogTextPlugin(interface.TextPlugin):
  """Text parser plugin for vsftpd log files."""

  NAME = 'vsftpd'
  DATA_FORMAT = 'vsftpd log file'

  _MONTH_DICT = {
      'jan': 1,
      'feb': 2,
      'mar': 3,
      'apr': 4,
      'may': 5,
      'jun': 6,
      'jul': 7,
      'aug': 8,
      'sep': 9,
      'oct': 10,
      'nov': 11,
      'dec': 12}

  _ONE_OR_TWO_DIGITS = pyparsing.Word(pyparsing.nums, max=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _THREE_LETTERS = pyparsing.Word(pyparsing.alphas, exact=3)

  # Date and time values are formatted as: Mon Jun  6 18:43:28 2016
  _DATE_TIME = pyparsing.Group(
      _THREE_LETTERS + _THREE_LETTERS + _ONE_OR_TWO_DIGITS +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') + _TWO_DIGITS +
      _FOUR_DIGITS)

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _LOG_LINE = (
      _DATE_TIME.set_results_name('date_time') +
      pyparsing.restOfLine().set_results_name('text') +
      _END_OF_LINE)

  _LINE_STRUCTURES = [('log_line', _LOG_LINE)]

  VERIFICATION_GRAMMAR = _LOG_LINE

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    event_data = VsftpdLogEventData()
    event_data.added_time = self._ParseTimeElements(time_elements_structure)
    # TODO: extract pid and username.
    event_data.text = self._GetStringValueFromStructure(structure, 'text')

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      _, month_string, day_of_month, hours, minutes, seconds, year = (
          time_elements_structure)

      month = self._MONTH_DICT.get(month_string.lower(), 0)

      time_elements_tuple = (year, month, day_of_month, hours, minutes, seconds)
      date_time = dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)
      date_time.is_local_time = True

      return date_time

    except (TypeError, ValueError) as exception:
      raise errors.ParseError(
          'Unable to parse time elements with error: {0!s}'.format(exception))

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugin(VsftpdLogTextPlugin)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the WinVer Windows Registry plugin."""

import unittest

from dfdatetime import filetime as dfdatetime_filetime
from dfwinreg import definitions as dfwinreg_definitions
from dfwinreg import fake as dfwinreg_fake

from plaso.parsers.winreg_plugins import windows_version

from tests import test_lib as shared_test_lib
from tests.parsers.winreg_plugins import test_lib


class WindowsRegistryInstallationEventDataTest(shared_test_lib.BaseTestCase):
  """Tests for the Windows installation event data attribute container."""

  def testGetAttributeNames(self):
    """Tests the GetAttributeNames function."""
    attribute_container = windows_version.WindowsRegistryInstallationEventData()

    expected_attribute_names = [
        '_event_data_stream_identifier',
        '_event_values_hash',
        '_parser_chain',
        'build_number',
        'data_type',
        'installation_time',
        'key_path',
        'owner',
        'product_name',
        'service_pack',
        'version']

    attribute_names = sorted(attribute_container.GetAttributeNames())

    self.assertEqual(attribute_names, expected_attribute_names)


class WindowsVersionPluginTest(test_lib.RegistryPluginTestCase):
  """Tests for the Windows version Windows Registry plugin."""

  def _CreateTestKey(self):
    """Creates Registry keys and values for testing.

    Returns:
      dfwinreg.WinRegistryKey: a Windows Registry key.
    """
    filetime = dfdatetime_filetime.Filetime()
    filetime.CopyFromDateTimeString('2012-08-31 20:09:55.123521')
    registry_key = dfwinreg_fake.FakeWinRegistryKey(
        'CurrentVersion', key_path_prefix='HKEY_LOCAL_MACHINE\\Software',
        last_written_time=filetime.timestamp, offset=153,
        relative_key_path='Microsoft\\Windows NT\\CurrentVersion')

    value_data = 'Service Pack 1'.encode('utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        'CSDVersion', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ, offset=1892)
    registry_key.AddValue(registry_value)

    value_data = '5.1'.encode('utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        'CurrentVersion', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ, offset=1121)
    registry_key.AddValue(registry_value)

    value_data = b'\x13\x1aAP'
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        'InstallDate', data=value_data,
        data_type=dfwinreg_definitions.REG_DWORD_LITTLE_ENDIAN, offset=1001)
    registry_key.AddValue(registry_value)

    value_data = 'MyTestOS'.encode('utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        'ProductName', data=value_data, data_type=dfwinreg_definitions.REG_SZ,
        offset=123)
    registry_key.AddValue(registry_value)

    value_data = 'A Concerned Citizen'.encode('utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        'RegisteredOwner', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ, offset=612)
    registry_key.AddValue(registry_value)

    return registry_key

  def testFilters(self):
    """Tests the FILTERS class attribute."""
    plugin = windows_version.WindowsVersionPlugin()

    self._AssertFiltersOnKeyPath(plugin, 'HKEY_LOCAL_MACHINE\\Software', (
        'Microsoft\\Windows NT\\CurrentVersion'))

    self._AssertNotFiltersOnKeyPath(
        plugin, 'HKEY_LOCAL_MACHINE\\Software', 'Bogus')

  def testProcess(self):
    """Tests the Process function."""
    registry_key = self._CreateTestKey()

    plugin = windows_version.WindowsVersionPlugin()
    storage_writer = self._ParseKeyWithPlugin(registry_key, plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 2)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_key_path = (
        'HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows NT\\CurrentVersion')

    expected_event_values = {
        'data_type': 'windows:registry:installation',
        'installation_time': '2012-08-31T20:09:55+00:00',
        'key_path': expected_key_path,
        'owner': 'A Concerned Citizen',
        'product_name': 'MyTestOS',
        'service_pack': 'Service Pack 1',
        'version': '5.1'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    expected_event_values = {
        'data_type': 'windows:registry:key_value',
        'key_path': expected_key_path,
        'last_written_time': '2012-08-31T20:09:55.1235210+00:00',
        'values': [
            ('CSDVersion', 'REG_SZ', 'Service Pack 1'),
            ('CurrentVersion', 'REG_SZ', '5.1'),
            ('ProductName', 'REG_SZ', 'MyTestOS'),
            ('RegisteredOwner', 'REG_SZ', 'A Concerned Citizen')]}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 1)
    self.CheckEventData(event_data, expected_event_values)

  def testProcessFile(self):
    """Tests the Process function on a Windows Registry file."""
    test_file_entry = self._GetTestFileEntry(['SOFTWARE-RunTests'])
    key_path = (
        'HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows NT\\CurrentVersion')

    win_registry = self._GetWinRegistryFromFileEntry(test_file_entry)
    registry_key = win_registry.GetKeyByPath(key_path)

    plugin = windows_version.WindowsVersionPlugin()
    storage_writer = self._ParseKeyWithPlugin(
        registry_key, plugin, file_entry=test_file_entry)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 2)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'windows:registry:installation',
        'installation_time': '2010-11-10T17:10:57+00:00',
        'key_path': key_path,
        'owner': 'Windows User',
        'product_name': 'Windows 7 Ultimate',
        'service_pack': 'Service Pack 1',
        'version': '6.1'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the archives CLI arguments helper."""

import argparse
import unittest

from plaso.cli import tools
from plaso.cli.helpers import archives
from plaso.lib import errors

from tests.cli import test_lib as cli_test_lib


class ArchivesArgumentsHelperTest(cli_test_lib.CLIToolTestCase):
  """Tests for the archives CLI arguments helper."""

  # pylint: disable=no-member,protected-access

  _EXPECTED_OUTPUT = """\
usage: cli_helper.py [--archives TYPES]

Test argument parser.

{0:s}:
  --archives TYPES  Define a list of archive and storage media image types for
                    which to process embedded file entries, such as TAR
                    (archive.tar) or ZIP (archive.zip). This is a comma
                    separated list where each entry is the name of an archive
                    type, such as "tar,zip". "all" indicates that all archive
                    types should be enabled. "none" disables processing file
                    entries embedded in archives. Use "--archives list" to
                    list the available archive types. WARNING: this can make
                    processing significantly slower.
""".format(cli_test_lib.ARGPARSE_OPTIONS)

  def testAddArguments(self):
    """Tests the AddArguments function."""
    argument_parser = argparse.ArgumentParser(
        prog='cli_helper.py', description='Test argument parser.',
        add_help=False,
        formatter_class=cli_test_lib.SortedArgumentsHelpFormatter)

    archives.ArchivesArgumentsHelper.AddArguments(argument_parser)

    output = self._RunArgparseFormatHelp(argument_parser)
    self.assertEqual(output, self._EXPECTED_OUTPUT)

  def testParseOptions(self):
    """Tests the ParseOptions function."""
    options = cli_test_lib.TestOptions()
    options.archives = 'tar'

    test_tool = tools.CLITool()
    archives.ArchivesArgumentsHelper.ParseOptions(options, test_tool)

    self.assertEqual(test_tool._archive_types_string, options.archives)

    with self.assertRaises(errors.BadConfigObject):
      archives.ArchivesArgumentsHelper.ParseOptions(options, None)

    with self.assertRaises(errors.BadConfigOption):
      options.archives = 'bogus'
      archives.ArchivesArgumentsHelper.ParseOptions(options, test_tool)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""This file contains the output module interface class."""

import abc

from plaso.output import logger


class OutputModule(object):
  """Output module interface."""

  NAME = ''
  DESCRIPTION = ''

  # Value to indicate the output module supports outputting additional fields.
  SUPPORTS_ADDITIONAL_FIELDS = False

  # Value to indicate the output module supports outputting custom fields.
  SUPPORTS_CUSTOM_FIELDS = False

  # Value to indicate the output module writes to an output file.
  WRITES_OUTPUT_FILE = False

  def _ReportEventError(self, event, event_data, error_message):
    """Reports an event related error.

    Args:
      event (EventObject): event.
      event_data (EventData): event data.
      error_message (str): error message.
    """
    event_identifier = event.GetIdentifier()
    event_identifier_string = event_identifier.CopyToString()
    display_name = getattr(event_data, 'display_name', None) or 'N/A'
    parser_chain = getattr(event_data, '_parser_chain', None) or 'N/A'

    error_message = (
        'Event: {0!s} data type: {1:s} display name: {2:s} '
        'parser chain: {3:s} with error: {4:s}').format(
            event_identifier_string, event_data.data_type, display_name,
            parser_chain, error_message)
    logger.error(error_message)

  def Close(self):
    """Closes the output."""
    return

  @abc.abstractmethod
  def GetFieldValues(
      self, output_mediator, event, event_data, event_data_stream, event_tag):
    """Retrieves the output field values.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
      event_tag (EventTag): event tag.

    Returns:
      dict[str, str]: output field values per name.
    """

  def GetMissingArguments(self):
    """Retrieves arguments required by the module that have not been specified.

    Returns:
      list[str]: names of argument that are required by the module and have
          not been specified.
    """
    return []

  def Open(self, **kwargs):  # pylint: disable=unused-argument
    """Opens the output."""
    return

  @abc.abstractmethod
  def WriteFieldValues(self, output_mediator, field_values):
    """Writes field values to the output.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      field_values (dict[str, str]): output field values per name.
    """

  def WriteFieldValuesOfMACBGroup(self, output_mediator, macb_group):
    """Writes field values of a MACB group to the output.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      macb_group (list[tuple[event, event_data, event_data_stream, event_tag]]):
          group of event, event_data, event_data_stream and event_tag objects
          with identical timestamps, attributes and values.
    """
    for event, event_data, event_data_stream, event_tag in macb_group:
      field_values = self.GetFieldValues(
          output_mediator, event, event_data, event_data_stream, event_tag)
      self.WriteFieldValues(output_mediator, field_values)

  def WriteFooter(self):
    """Writes the footer to the output.

    Can be used for post-processing or output after the last event is written,
    such as writing a file footer.
    """
    return

  def WriteHeader(self, output_mediator):  # pylint: disable=unused-argument
    """Writes the header to the output.

    Can be used for pre-processing or output before the first event
    is written, such as writing a file header.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
    """
    return



# -*- coding: utf-8 -*-
"""Event attribute containers."""

import hashlib
import re

from acstore.containers import interface
from acstore.containers import manager

from dfdatetime import interface as dfdatetime_interface


def CalculateEventValuesHash(event_data, event_data_stream):
  """Calculates a digest hash of the event values.

  Args:
    event_data (EventData): event data.
    event_data_stream (EventDataStream): an event data stream or None if not
        available.

  Returns:
    str: digest hash of the event values content.

  Raises:
    RuntimeError: if the event values hash cannot be determined.
  """
  attributes = ['data_type: {0:s}'.format(event_data.data_type)]

  for attribute_name, attribute_value in sorted(event_data.GetAttributes()):
    if attribute_value is None or attribute_name in (
        '_event_data_stream_identifier', '_event_values_hash', '_parser_chain',
        'data_type'):
      continue

    # Ignore date and time values.
    if isinstance(attribute_value, dfdatetime_interface.DateTimeValues):
      continue

    if (isinstance(attribute_value, list) and attribute_value and
        isinstance(attribute_value[0],
                     dfdatetime_interface.DateTimeValues)):
      continue

    if not isinstance(attribute_value, (bool, float, int, list, str)):
      raise RuntimeError(
          'Unsupported attribute: {0:s} value type: {1!s}'.format(
              attribute_name, type(attribute_value)))

    try:
      attribute_string = '{0:s}: {1!s}'.format(
          attribute_name, attribute_value)
      attributes.append(attribute_string)
    except UnicodeDecodeError:
      raise RuntimeError(
          'Failed to decode attribute {0:s}'.format(attribute_name))

  if event_data_stream:
    for attribute_name, attribute_value in sorted(
        event_data_stream.GetAttributes()):

      if attribute_name == 'path_spec':
        attribute_value = attribute_value.comparable

      elif not isinstance(attribute_value, (bool, float, int, list, str)):
        raise RuntimeError(
            'Unsupported attribute: {0:s} value type: {1!s}'.format(
                attribute_name, type(attribute_value)))

      try:
        attribute_string = '{0:s}: {1!s}'.format(
            attribute_name, attribute_value)
        attributes.append(attribute_string)
      except UnicodeDecodeError:
        raise RuntimeError(
            'Failed to decode attribute {0:s}'.format(attribute_name))

  content = ', '.join(attributes)
  content_data = content.encode('utf-8')

  md5_context = hashlib.md5(content_data)

  return md5_context.hexdigest()


class DateLessLogHelper(interface.AttributeContainer):
  """Attribute container to assist with logs without full dates.

  Attributes:
    earliest_date (list[int, int, int]): earliest possible date the event data
        stream was created. The date is a tuple of year, month and day of month.
    granularity (str): granularity of the date-less log format.
    last_relative_date (list[int, int, int]): last relative date determined by
        the date-less log helper. The date is a tuple of year, month and day of
        month.
    latest_date (List[int]): latest possible date the event data stream was
        created. The date is a tuple of year, month and day of month.
  """

  CONTAINER_TYPE = 'date_less_log_helper'

  SCHEMA = {
      '_event_data_stream_identifier': 'AttributeContainerIdentifier',
      'earliest_date': 'List[int]',
      'granularity': 'str',
      'last_relative_date': 'List[int]',
      'latest_date': 'List[int]'}

  _SERIALIZABLE_PROTECTED_ATTRIBUTES = [
      '_event_data_stream_identifier']

  # The date-less log format only supports time.
  GRANULARITY_NO_DATE = 'd'

  # The date-less log format only supports month and day of month.
  GRANULARITY_NO_YEAR = 'y'

  def __init__(self):
    """Initializes a date-less log helper attribute container."""
    super(DateLessLogHelper, self).__init__()
    self._event_data_stream_identifier = None
    self.earliest_date = None
    self.granularity = self.GRANULARITY_NO_YEAR
    self.last_relative_date = None
    self.latest_date = None

  # TODO: the YearLessLogHelper attribute container is kept for backwards
  # compatibility remove once storage format 20230327 is obsolete.
  def CopyFromYearLessLogHelper(self, year_less_log_helper):
    """Copy the values of a year-less log helper.

    Args:
      year_less_log_helper (YearLessLogHelper): year-less log helper.
    """
    self.earliest_date = (year_less_log_helper.earliest_year, 1, 1)
    self.granularity = self.GRANULARITY_NO_YEAR
    self.last_relative_date = (year_less_log_helper.last_relative_year, 0, 0)
    self.latest_date = (year_less_log_helper.latest_year, 1, 1)

  def GetEarliestDate(self):
    """Retrieves the earliest date adjusted to the granularity.

    Returns:
      tuple[int, int, int]: earliest date as tuple of year, month and day of
          month or None if not available.
    """
    if self.earliest_date and self.granularity == self.GRANULARITY_NO_YEAR:
      return self.earliest_date[0], 0, 0

    return self.earliest_date

  def GetEventDataStreamIdentifier(self):
    """Retrieves the identifier of the associated event data stream.

    The event data stream identifier is a storage specific value that requires
    special handling during serialization.

    Returns:
      AttributeContainerIdentifier: event data stream or None when not set.
    """
    return self._event_data_stream_identifier

  def GetLastRelativeDate(self):
    """Retrieves the last relative date adjusted to the granularity.

    Returns:
      tuple[int, int, int]: last relative date as tuple of year, month and day
          of month or None if not available.
    """
    if (self.last_relative_date and
        self.granularity == self.GRANULARITY_NO_YEAR):
      return self.last_relative_date[0], 0, 0

    return self.last_relative_date

  def GetLatestDate(self):
    """Retrieves the latest date adjusted to the granularity.

    Returns:
      tuple[int, int, int]: latest date as tuple of year, month and day of
          month or None if not available.
    """
    if self.latest_date and self.granularity == self.GRANULARITY_NO_YEAR:
      return self.latest_date[0], 0, 0

    return self.latest_date

  def SetEventDataStreamIdentifier(self, event_data_stream_identifier):
    """Sets the identifier of the associated event data stream.

    The event data stream identifier is a storage specific value that requires
    special handling during serialization.

    Args:
      event_data_stream_identifier (AttributeContainerIdentifier): event data
          stream identifier.
    """
    self._event_data_stream_identifier = event_data_stream_identifier


class EventData(interface.AttributeContainer):
  """Event data attribute container.

  The event data attribute container represents the attributes of an entity,
  such as a database record or log line.

  Attributes:
    data_type (str): event data type indicator.
  """

  CONTAINER_TYPE = 'event_data'

  _SERIALIZABLE_PROTECTED_ATTRIBUTES = [
      '_event_data_stream_identifier',
      '_event_values_hash',
      '_parser_chain']

  def __init__(self, data_type=None):
    """Initializes an event data attribute container.

    Args:
      data_type (Optional[str]): event data type indicator.
    """
    super(EventData, self).__init__()
    self._event_data_stream_identifier = None
    self._event_values_hash = None
    self._parser_chain = None

    self.data_type = data_type

  def GetAttributeValuesString(self):
    """Retrieves a comparable string of the attribute values.

    Returns:
      str: comparable string of the attribute values.

    Raises:
      TypeError: if the attribute value type is not supported.
    """
    attributes = []
    for attribute_name, attribute_value in sorted(self.__dict__.items()):
      # Not using startswith to improve performance.
      if attribute_name[0] == '_' or attribute_value is None:
        continue

      if isinstance(attribute_value, bytes):
        raise TypeError(
            'Attribute: {0:s} value of type bytes not supported.'.format(
                attribute_name))

      if isinstance(attribute_value, dict):
        raise TypeError(
            'Attribute: {0:s} value of type dict not supported.'.format(
                attribute_name))

      attribute_string = '{0:s}: {1!s}'.format(attribute_name, attribute_value)
      attributes.append(attribute_string)

    return ', '.join(attributes)

  def GetEventDataStreamIdentifier(self):
    """Retrieves the identifier of the associated event data stream.

    The event data stream identifier is a storage specific value that requires
    special handling during serialization.

    Returns:
      AttributeContainerIdentifier: event data stream or None when not set.
    """
    return self._event_data_stream_identifier

  def SetEventDataStreamIdentifier(self, event_data_stream_identifier):
    """Sets the identifier of the associated event data stream.

    The event data stream identifier is a storage specific value that requires
    special handling during serialization.

    Args:
      event_data_stream_identifier (AttributeContainerIdentifier): event data
          stream identifier.
    """
    self._event_data_stream_identifier = event_data_stream_identifier


class EventDataStream(interface.AttributeContainer):
  """Event data stream attribute container.

  The event data stream attribute container represents the attributes of
  a data stream, such as the content of a file or extended attribute.

  Attributes:
    file_entropy (str): byte entropy value of the data stream.
    md5_hash (str): MD5 digest hash of the data stream.
    path_spec (dfvfs.PathSpec): path specification of the data stream.
    sha1_hash (str): SHA-1 digest hash of the data stream.
    sha256_hash (str): SHA-256 digest hash of the data stream.
    yara_match (list[str]): names of the Yara rules that matched the data
        stream.
  """

  CONTAINER_TYPE = 'event_data_stream'

  SCHEMA = {
      'file_entropy': 'str',
      'md5_hash': 'str',
      'path_spec': 'dfvfs.PathSpec',
      'sha1_hash': 'str',
      'sha256_hash': 'str',
      'yara_match': 'List[str]'}

  def __init__(self):
    """Initializes an event data attribute container."""
    super(EventDataStream, self).__init__()
    self.file_entropy = None
    self.md5_hash = None
    self.path_spec = None
    self.sha1_hash = None
    self.sha256_hash = None
    self.yara_match = None


class EventObject(interface.AttributeContainer):
  """Event attribute container.

  The framework is designed to parse files and create events
  from individual records, log lines or keys extracted from files.
  The event object provides an extensible data store for event
  attributes.

  Attributes:
    date_time (dfdatetime.DateTimeValues): date and time values.
    timestamp (int): timestamp, which contains the number of microseconds
        since January 1, 1970, 00:00:00 UTC.
    timestamp_desc (str): description of the meaning of the timestamp.
  """

  CONTAINER_TYPE = 'event'

  SCHEMA = {
      '_event_data_identifier': 'AttributeContainerIdentifier',
      'date_time': 'dfdatetime.DateTimeValues',
      'timestamp': 'int',
      'timestamp_desc': 'str'}

  _SERIALIZABLE_PROTECTED_ATTRIBUTES = [
      '_event_data_identifier']

  def __init__(self):
    """Initializes an event attribute container."""
    super(EventObject, self).__init__()
    self._event_data_identifier = None
    self.date_time = None
    self.timestamp = None
    # TODO: rename timestamp_desc to timestamp_description
    self.timestamp_desc = None

  # This method is necessary for heap sort.
  def __lt__(self, other):
    """Compares if the event attribute container is less than the other.

    Events are compared by timestamp.

    Args:
      other (EventObject): event attribute container to compare to.

    Returns:
      bool: True if the event attribute container is less than the other.
    """
    return (self.timestamp < other.timestamp or
            self.timestamp_desc < other.timestamp_desc)

  def GetEventDataIdentifier(self):
    """Retrieves the identifier of the associated event data.

    The event data identifier is a storage specific value that requires special
    handling during serialization.

    Returns:
      AttributeContainerIdentifier: event data identifier or None when not set.
    """
    return self._event_data_identifier

  def SetEventDataIdentifier(self, event_data_identifier):
    """Sets the identifier of the associated event data.

    The event data identifier is a storage specific value that requires special
    handling during serialization.

    Args:
      event_data_identifier (AttributeContainerIdentifier): event data
          identifier.
    """
    self._event_data_identifier = event_data_identifier


class EventTag(interface.AttributeContainer):
  """Event tag attribute container.

  Attributes:
    labels (list[str]): labels, such as "malware", "application_execution".
  """

  CONTAINER_TYPE = 'event_tag'

  SCHEMA = {
      '_event_identifier': 'AttributeContainerIdentifier',
      'labels': 'List[str]'}

  _INVALID_LABEL_CHARACTERS_REGEX = re.compile(r'[^A-Za-z0-9_]')

  _SERIALIZABLE_PROTECTED_ATTRIBUTES = [
      '_event_identifier']

  _VALID_LABEL_REGEX = re.compile(r'^[A-Za-z0-9_]+$')

  def __init__(self):
    """Initializes an event tag attribute container."""
    super(EventTag, self).__init__()
    self._event_identifier = None
    self.labels = []

  def AddLabel(self, label):
    """Adds a label to the event tag.

    Args:
      label (str): label.

    Raises:
      TypeError: if the label provided is not a string.
      ValueError: if a label is malformed.
    """
    if not isinstance(label, str):
      raise TypeError('label is not a string type. Is {0!s}'.format(
          type(label)))

    if not self._VALID_LABEL_REGEX.match(label):
      raise ValueError((
          'Unsupported label: "{0:s}". A label must only consist of '
          'alphanumeric characters or underscores.').format(label))

    if label not in self.labels:
      self.labels.append(label)

  def AddLabels(self, labels):
    """Adds labels to the event tag.

    Args:
      labels (list[str]): labels.

    Raises:
      ValueError: if a label is malformed.
    """
    for label in labels:
      if not label or not self._VALID_LABEL_REGEX.match(label):
        raise ValueError((
            'Unsupported label: "{0!s}". A label must only consist of '
            'alphanumeric characters or underscores.').format(label))

    for label in labels:
      if label not in self.labels:
        self.labels.append(label)

  @classmethod
  def CopyTextToLabel(cls, text, prefix=''):
    """Copies a string to a label.

    A label only supports a limited set of characters therefore
    unsupported characters are replaced with an underscore.

    Args:
      text (str): label text.
      prefix (Optional[str]): label prefix.

    Returns:
      str: label.
    """
    text = '{0:s}{1:s}'.format(prefix, text)
    return cls._INVALID_LABEL_CHARACTERS_REGEX.sub('_', text)

  def GetEventIdentifier(self):
    """Retrieves the identifier of the associated event.

    The event identifier is a storage specific value that requires special
    handling during serialization.

    Returns:
      AttributeContainerIdentifier: event identifier or None when not set.
    """
    return self._event_identifier

  def SetEventIdentifier(self, event_identifier):
    """Sets the identifier of the associated event.

    The event identifier is a storage specific value that requires special
    handling during serialization.

    Args:
      event_identifier (AttributeContainerIdentifier): event identifier.
    """
    self._event_identifier = event_identifier


# TODO: the YearLessLogHelper attribute container is kept for backwards
# compatibility remove once storage format 20230327 is obsolete.
class YearLessLogHelper(interface.AttributeContainer):
  """Year-less log helper attribute container.

  Attributes:
    earliest_year (int): earliest possible year the event data stream was
        created.
    last_relative_year (int): last relative year determined by the year-less
        log helper.
    latest_year (int): latest possible year the event data stream was created.
  """

  CONTAINER_TYPE = 'year_less_log_helper'

  SCHEMA = {
      '_event_data_stream_identifier': 'AttributeContainerIdentifier',
      'earliest_year': 'int',
      'last_relative_year': 'int',
      'latest_year': 'int'}

  _SERIALIZABLE_PROTECTED_ATTRIBUTES = [
      '_event_data_stream_identifier']

  def __init__(self):
    """Initializes a year-less log helper attribute container."""
    super(YearLessLogHelper, self).__init__()
    self._event_data_stream_identifier = None
    self.earliest_year = None
    self.last_relative_year = None
    self.latest_year = None

  def GetEventDataStreamIdentifier(self):
    """Retrieves the identifier of the associated event data stream.

    The event data stream identifier is a storage specific value that requires
    special handling during serialization.

    Returns:
      AttributeContainerIdentifier: event data stream or None when not set.
    """
    return self._event_data_stream_identifier

  def SetEventDataStreamIdentifier(self, event_data_stream_identifier):
    """Sets the identifier of the associated event data stream.

    The event data stream identifier is a storage specific value that requires
    special handling during serialization.

    Args:
      event_data_stream_identifier (AttributeContainerIdentifier): event data
          stream identifier.
    """
    self._event_data_stream_identifier = event_data_stream_identifier


manager.AttributeContainersManager.RegisterAttributeContainers([
    DateLessLogHelper, EventData, EventDataStream, EventObject, EventTag,
    YearLessLogHelper])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Windows shell folders helper."""

import unittest

from plaso.helpers.windows import shell_folders

from tests import test_lib as shared_test_lib


class WindowsShellFoldersHelperTest(shared_test_lib.BaseTestCase):
  """Tests for the Windows shell folders helper."""

  def testGetDescription(self):
    """Tests the GetDescription function."""
    description = shell_folders.WindowsShellFoldersHelper.GetDescription(
        '3080f90d-d7ad-11d9-bd98-0000947b0257')
    self.assertEqual(description, 'Show Desktop')

    description = shell_folders.WindowsShellFoldersHelper.GetDescription(
        'bogus')
    self.assertIsNone(description)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Text parser plugin for Google-formatted log files.

Note that this format is also used by Kubernetes.

Also see:
  https://github.com/google/glog
  https://github.com/kubernetes/klog
"""

import re

from dfdatetime import time_elements as dfdatetime_time_elements

import pyparsing

from plaso.containers import events
from plaso.lib import dateless_helper
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class GoogleLogEventData(events.EventData):
  """Google-formatted log file event data.

  Attributes:
    file_name (str): the name of the source file that logged the message.
    last_written_time (dfdatetime.DateTimeValues): entry last written date and
        time.
    line_number (int): the line number in the source file where the logging
        statement is.
    message (str): the log message.
    priority (str): the priority of the message - I, W, E or F. These values
        represent messages logged at INFO, WARNING, ERROR or FATAL severities,
        respectively.
    thread_identifier (int): the identifier of the thread that recorded the
        message.
  """

  DATA_TYPE = 'googlelog:log'

  def __init__(self, data_type=DATA_TYPE):
    """Initializes event data.

    Args:
      data_type (Optional[str]): event data type indicator.
    """
    super(GoogleLogEventData, self).__init__(data_type=data_type)
    self.file_name = None
    self.last_written_time = None
    self.line_number = None
    self.message = None
    self.priority = None
    self.thread_identifier = None


class GoogleLogTextPlugin(
    interface.TextPlugin, dateless_helper.DateLessLogFormatHelper):
  """Text parser plugin for Google-formatted log files."""

  NAME = 'googlelog'
  DATA_FORMAT = 'Google-formatted log file'

  _ONE_OR_TWO_DIGITS = pyparsing.Word(pyparsing.nums, max=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _SIX_DIGITS = pyparsing.Word(pyparsing.nums, exact=6).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  # Date and time values are formatted as: MMDD hh:mm:ss.######
  # For example: 1231 23:59:59.000001
  _DATE_TIME = (
      _TWO_DIGITS + _ONE_OR_TWO_DIGITS +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Optional(
          pyparsing.Suppress('.') + _SIX_DIGITS)).set_results_name('date_time')

  _PRIORITY = pyparsing.one_of(['E', 'F', 'I', 'W']).set_results_name(
      'priority')

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _LOG_LINE = (
      _PRIORITY + _DATE_TIME +
      pyparsing.Word(pyparsing.nums).set_results_name('thread_identifier') +
      pyparsing.Word(pyparsing.printables.replace(':', '')).set_results_name(
          'file_name') + pyparsing.Suppress(':') +
      pyparsing.Word(pyparsing.nums).set_results_name('line_number') +
      pyparsing.Suppress('] ') +
      pyparsing.Regex(
          '.*?(?=($|\n[IWEF][0-9]{4}))', re.DOTALL).set_results_name(
              'message') + _END_OF_LINE)

  # Header date and time values are formatted as: 2019/07/18 06:07:40
  _HEADER_DATE_TIME = (
      _FOUR_DIGITS + pyparsing.Suppress('/') +
      _TWO_DIGITS + pyparsing.Suppress('/') + _ONE_OR_TWO_DIGITS +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') + _TWO_DIGITS)

  _HEADER_LINES = (
      pyparsing.Regex(
          r'Log file created at: (?P<date_time>[0-9]{4}/[0-9]{2}/[0-9]{2} '
          r'[0-9]{2}:[0-9]{2}:[0-9]{2})\n') +
      pyparsing.Regex(r'Running on machine: .*\n'))

  _HEADER_GRAMMAR = (
      _HEADER_LINES +
      pyparsing.Optional(
          pyparsing.Regex(r'Application fingerprint: .*\n') ^
          pyparsing.Regex(r'Binary: .*\n')) +
      pyparsing.Optional(
          pyparsing.Regex(r'Running duration (h:mm:ss): .*\n')) +
      pyparsing.Regex(r'Log line format: (?P<log_line_format>.*)\n'))

  _LINE_STRUCTURES = [('log_line', _LOG_LINE)]

  VERIFICATION_GRAMMAR = _HEADER_LINES

  VERIFICATION_LITERALS = ['Log file created at: ', 'Running on machine: ']

  def _ParseHeader(self, parser_mediator, text_reader):
    """Parses a text-log file header.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Raises:
      ParseError: when the header cannot be parsed.
    """
    try:
      structure_generator = self._HEADER_GRAMMAR.scan_string(
          text_reader.lines, max_matches=1)
      structure, start, end = next(structure_generator)

    except StopIteration:
      structure = None

    except pyparsing.ParseException as exception:
      raise errors.ParseError(exception)

    if not structure or start != 0:
      raise errors.ParseError('No match found.')

    date_time_structure = self._GetValueFromStructure(structure, 'date_time')

    try:
      time_elements_structure = self._HEADER_DATE_TIME.parse_string(
          date_time_structure)
    except pyparsing.ParseException:
      raise errors.ParseError('Unable to parser header date time.')

    year, month, _, _, _, _ = time_elements_structure

    try:
      self._SetMonthAndYear(month, year)
    except (TypeError, ValueError):
      raise errors.ParseError('Invalid header date time value.')

    log_line_format = self._GetValueFromStructure(structure, 'log_line_format')
    if log_line_format != '[IWEF]mmdd hh:mm:ss.uuuuuu threadid file:line] msg':
      raise errors.ParseError('Unsupported log line format.')

    # TODO: create log file event data with creation time.

    text_reader.SkipAhead(end)

  def _ParseLine(self, parser_mediator, structure):
    """Process a single log line into a GoogleLogEvent.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): elements parsed from the file.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    event_data = GoogleLogEventData()
    event_data.file_name = self._GetValueFromStructure(structure, 'file_name')
    event_data.last_written_time = self._ParseTimeElements(
        time_elements_structure)
    event_data.line_number = self._GetValueFromStructure(
        structure, 'line_number')
    event_data.message = self._GetValueFromStructure(structure, 'message')
    event_data.priority = self._GetValueFromStructure(structure, 'priority')
    event_data.thread_identifier = self._GetValueFromStructure(
        structure, 'thread_identifier')

    parser_mediator.ProduceEventData(event_data)

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    self._ParseLine(parser_mediator, structure)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      month, day_of_month, hours, minutes, seconds, microseconds = (
          time_elements_structure)

      self._UpdateYear(month)

      relative_year = self._GetRelativeYear()

      time_elements_tuple = (
          relative_year, month, day_of_month, hours, minutes, seconds,
          microseconds)

      date_time = dfdatetime_time_elements.TimeElementsInMicroseconds(
          is_delta=True, time_elements_tuple=time_elements_tuple)

      date_time.is_local_time = True

      return date_time

    except (TypeError, ValueError) as exception:
      raise errors.ParseError(
          'Unable to parse time elements with error: {0!s}'.format(exception))

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    date_time_structure = self._GetValueFromStructure(structure, 'date_time')

    try:
      time_elements_structure = self._HEADER_DATE_TIME.parse_string(
          date_time_structure)
    except pyparsing.ParseException:
      return False

    try:
      year, month, day_of_month, hours, minutes, seconds = (
          time_elements_structure)

      time_elements_tuple = (year, month, day_of_month, hours, minutes, seconds)

      dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)

    except (TypeError, ValueError):
      return False

    self._SetEstimatedYear(parser_mediator)

    return True


text_parser.TextLogParser.RegisterPlugin(GoogleLogTextPlugin)
